@@ -49,9 +49,12 @@ from jax._src.config import (
   default_matmul_precision as default_matmul_precision,
   default_prng_impl as default_prng_impl,
   numpy_rank_promotion as numpy_rank_promotion,
+  jax2tf_associative_scan_reductions as jax2tf_associative_scan_reductions
 )
+from .core import eval_context as ensure_compile_time_eval
 from jax._src.api import (
   ad,  # TODO(phawkins): update users to avoid this.
+  block_until_ready,
   checkpoint as checkpoint,
   checkpoint_policies as checkpoint_policies,
   closure_convert as closure_convert,
