@@ -463,6 +463,31 @@ flags.DEFINE_bool(
         'Has no effect on TPU, since only the outfeed mechanism is implemented.'
     )
 )
+flags.DEFINE_bool(
+    'jax_host_callback_ad_transforms',
+    bool_env('JAX_HOST_CALLBACK_AD_TRANSFORMS', False),
+    help=(
+        'Enable support for jvp/vjp for the host_callback primitives. Default is '
+        'False, which means that host_callback operates only on primals. '
+        'The flag exists only temporarily, for backward compatibility.'
+    )
+)
+
+# TODO(b/214340779): remove flag when XLA:CPU is improved.
+jax2tf_associative_scan_reductions = config.define_bool_state(
+    name='jax2tf_associative_scan_reductions',
+    default=False,
+    help=(
+        'JAX has two separate lowering rules for the cumulative reduction '
+        'primitives (cumsum, cumprod, cummax, cummin). On CPUs and GPUs it uses '
+        'a lax.associative_scan, while for TPUs it uses the HLO ReduceWindow. '
+        'The latter has a slow implementation on CPUs and GPUs. '
+        'By default, jax2tf uses the TPU lowering. Set this flag to True to '
+        'use the associative scan lowering usage, and only if it makes a difference '
+        'for your application. '
+        'See the jax2tf README.md for more details.'
+    )
+)
 
 enable_checks = config.define_bool_state(
     name='jax_enable_checks',
@@ -503,8 +528,8 @@ log_compiles = config.define_bool_state(
           'option is set, the log level is WARNING; otherwise the level is '
           'DEBUG.'))
 
-gsda_out = config.define_bool_state(
-    name='jax_gsda_out',
+parallel_functions_output_gda = config.define_bool_state(
+    name='jax_parallel_functions_output_gda',
     default=False,
     help='If True, pjit will output GSDAs.')
 
@@ -635,6 +660,6 @@ traceback_filtering = config.define_enum_state(
 
 enable_mlir = config.define_bool_state(
     name='jax_enable_mlir',
-    default=False,
+    default=lib.mlir_api_version >= 1,
     help=('Enables an experimental code path that compiles JAX programs via '
           'emitting the MLIR MHLO dialect.'))
