@@ -14,11 +14,15 @@
 
 # Primitive dispatch and jit dispatch.
 
+import contextlib
 from functools import partial
 import itertools
+import time
 from typing import (
     Any, Callable, Dict, Optional, Sequence, Set, Tuple, Type, Union)
 from typing_extensions import Protocol
+import os
+import re
 import warnings
 
 from absl import logging
@@ -34,15 +38,25 @@ import jax.interpreters.xla as xla
 import jax.interpreters.partial_eval as pe
 from jax.errors import UnexpectedTracerError
 from jax._src.abstract_arrays import array_types
-from jax._src.config import config
+from jax._src.config import config, flags
 from jax._src import device_array
 from jax._src import dtypes
 from jax._src import profiler
+from jax._src.lib.mlir import ir
 from jax._src.lib import xla_bridge as xb
 from jax._src.lib import xla_client as xc
 import jax._src.util as util
 from jax._src import traceback_util
 
+FLAGS = flags.FLAGS
+
+flags.DEFINE_string(
+    'jax_dump_ir_to', os.getenv('JAX_DUMP_IR_TO', ''),
+    help="Path to which HLO/MHLO IR that is emitted by JAX as input to the "
+         "compiler should be dumped as text files. Optional. If omitted, JAX "
+         "will not dump IR.")
+
+
 traceback_util.register_exclusion(__file__)
 
 MYPY = False  # Are we currently type checking with mypy?
@@ -158,6 +172,18 @@ def _xla_callable_uncached(fun: lu.WrappedFun, device, backend, name,
 _xla_callable = lu.cache(_xla_callable_uncached)
 
 
+@contextlib.contextmanager
+def log_elapsed_time(fmt: str):
+  if _on_exit:
+    yield
+  else:
+    log_priority = logging.WARNING if config.jax_log_compiles else logging.DEBUG
+    start_time = time.time()
+    yield
+    elapsed_time = time.time() - start_time
+    logging.log(log_priority, fmt.format(elapsed_time=elapsed_time))
+
+
 @profiler.annotate_function
 def lower_xla_callable(fun: lu.WrappedFun, device, backend, name,
                        donated_invars, *arg_specs):
@@ -166,8 +192,10 @@ def lower_xla_callable(fun: lu.WrappedFun, device, backend, name,
                      "got device={} and backend={}".format(device, backend))
 
   abstract_args, arg_devices = util.unzip2(arg_specs)
-  jaxpr, out_avals, consts = pe.trace_to_jaxpr_final(
-      fun, abstract_args, pe.debug_info_final(fun, "jit"))
+  with log_elapsed_time(f"Finished tracing + transforming {fun.__name__} "
+                        "for jit in {elapsed_time} sec"):
+    jaxpr, out_avals, consts = pe.trace_to_jaxpr_final(
+        fun, abstract_args, pe.debug_info_final(fun, "jit"))
   if any(isinstance(c, core.Tracer) for c in consts):
     raise UnexpectedTracerError("Encountered an unexpected tracer.")
   jaxpr, kept_const_idx, kept_var_idx = _prune_unused_inputs(jaxpr)
@@ -189,13 +217,16 @@ def lower_xla_callable(fun: lu.WrappedFun, device, backend, name,
   # and don't need to evaluate their arguments.
   if not jaxpr.eqns:
     return XlaComputation(
-        name, None, True, None, jaxpr, consts, device, abstract_args, out_avals,
-        kept_var_idx)
+        name, None, True, None, jaxpr=jaxpr, consts=consts, device=device,
+        in_avals=abstract_args, out_avals=out_avals, kept_var_idx=kept_var_idx)
 
   if not _on_exit:
     log_priority = logging.WARNING if config.jax_log_compiles else logging.DEBUG
-    logging.log(log_priority, "Compiling %s (%s) for args %s.",
-                fun.__name__, id(fun), abstract_args)
+    if len(abstract_args) > 10:
+      msg = f"Compiling {fun.__name__} ({id(fun)}) for {len(abstract_args)} args."
+    else:
+      msg = f"Compiling {fun.__name__} ({id(fun)} for args {abstract_args}."
+    logging.log(log_priority, msg)
 
   if nreps > 1:
     warnings.warn(
@@ -222,17 +253,20 @@ def lower_xla_callable(fun: lu.WrappedFun, device, backend, name,
   name_stack = xla.extend_name_stack(xla.wrap_name(name, 'jit'))
   closed_jaxpr = core.ClosedJaxpr(jaxpr, consts)
   module: Union[str, xc.XlaComputation]
+  module_name = f"jit_{fun.__name__}"
   if config.jax_enable_mlir:
     module = mlir.lower_jaxpr_to_module(
-        closed_jaxpr, backend.platform, axis_env, name_stack, donated_invars)
+        module_name, closed_jaxpr, backend.platform, axis_env, name_stack,
+        donated_invars)
   else:
     module = xla.lower_jaxpr_to_xla_module(
-        f"jit_{fun.__name__}", closed_jaxpr, backend.platform, axis_env,
+        module_name, closed_jaxpr, backend.platform, axis_env,
         name_stack, tuple_args, donated_invars, replicated_args=None,
         arg_partitions=None, out_partitions=None)
   return XlaComputation(
-      name, module, False, donated_invars, nreps, device, backend, tuple_args,
-      abstract_args, out_avals, kept_var_idx)
+      name, module, False, donated_invars, nreps=nreps, device=device,
+      backend=backend, tuple_args=tuple_args, in_avals=abstract_args,
+      out_avals=out_avals, kept_var_idx=kept_var_idx)
 
 
 def prefetch(x):
@@ -287,32 +321,6 @@ def apply_outfeed_rewriter(jaxpr: core.Jaxpr) -> core.Jaxpr:
   else:
     return jaxpr
 
-outfeed_primitives: Set[core.Primitive] = set()
-def jaxpr_uses_outfeed(jaxpr: core.Jaxpr) -> bool:
-  """Finds if there are outfeed primitives anywhere inside a Jaxpr."""
-  return any(primitive_uses_outfeed(eqn.primitive, eqn.params)
-             for eqn in jaxpr.eqns)
-
-def _param_uses_outfeed(param):
-  if type(param) is core.Jaxpr:
-    if jaxpr_uses_outfeed(param):
-      return True
-  elif type(param) is core.ClosedJaxpr:
-    if jaxpr_uses_outfeed(param.jaxpr):
-      return True
-  return False
-
-def primitive_uses_outfeed(prim: core.Primitive, params: Dict) -> bool:
-  if prim in outfeed_primitives:
-    return True
-  for param in params.values():
-    if isinstance(param, tuple):
-      if any(unsafe_map(_param_uses_outfeed, param)):
-        return True
-    elif _param_uses_outfeed(param):
-      return True
-  return False
-
 
 def jaxpr_replicas(jaxpr) -> int:
   """The number of replicas needed for a jaxpr.
@@ -481,7 +489,8 @@ class XlaComputation:
   _donated_invars: Optional[Sequence[bool]]
 
   def __init__(self, name: str, hlo, is_trivial: bool,
-               donated_invars: Optional[Sequence[bool]], *compile_args):
+               donated_invars: Optional[Sequence[bool]],
+               **compile_args):
     self.name = name
     self._hlo = hlo
     self._is_trivial = is_trivial
@@ -492,19 +501,31 @@ class XlaComputation:
   def is_trivial(self):
     return self._is_trivial
 
-  def hlo(self):
+  def hlo(self) -> xc.XlaComputation:
     if self.is_trivial():
       raise ValueError("A trivial computation has no HLO")
-    return self._hlo
+    if isinstance(self._hlo, xc.XlaComputation):
+      return self._hlo
+    return xe.mlir.mlir_module_to_xla_computation(
+        mlir.module_to_string(self._hlo),
+        use_tuple_args=self.compile_args["tuple_args"])
+
+  def mhlo(self) -> str:
+    if self.is_trivial():
+      raise ValueError("A trivial computation has no MHLO")
+    if isinstance(self._hlo, xc.XlaComputation):
+      return xe.mlir.xla_computation_to_mlir_module(self._hlo)
+    return mlir.module_to_string(self._hlo)
 
   def compile(self) -> 'XlaCompiledComputation':
     if self._executable is None:
       if self.is_trivial():
         self._executable = XlaCompiledComputation.from_trivial_jaxpr(
-            *self.compile_args)
+            **self.compile_args)
       else:
         self._executable = XlaCompiledComputation.from_xla_computation(
-            self.name, self.hlo(), *self.compile_args)
+            self.name, self._hlo, **self.compile_args)
+
     return self._executable
 
 @profiler.annotate_function
@@ -516,22 +537,47 @@ def backend_compile(backend, built_c, options):
 # TODO(phawkins): update users.
 xla.backend_compile = backend_compile
 
+_ir_dump_counter = itertools.count()
+
+def _make_string_safe_for_filename(s: str) -> str:
+  return re.sub(r'[^\w.)( -]', '', s)
+
+def _dump_ir_to_file(name: str, ir: str):
+  id = next(_ir_dump_counter)
+  name = f"jax_ir{id}_{_make_string_safe_for_filename(name)}.mlir"
+  name = os.path.join(FLAGS.jax_dump_ir_to, name)
+  with open(name, "w") as f:
+    f.write(ir)
+
+
 def compile_or_get_cached(backend, computation, compile_options):
-    # Avoid import cycle between jax and jax.experimental
-    from jax.experimental.compilation_cache import compilation_cache as cc
-    # Persistent compilation cache only implemented on TPU.
-    # TODO(skye): add warning when initializing cache on unsupported default platform
-    if cc.is_initialized() and backend.platform == 'tpu':
-        cached_executable = cc.get_executable(computation, compile_options, backend)
-        if cached_executable is not None:
-            logging.info('Persistent compilation cache hit for %s.',
-                         computation.name())
-            return cached_executable
-        else:
-            compiled = backend_compile(backend, computation, compile_options)
-            cc.put_executable(computation, compile_options, compiled, backend)
-            return compiled
-    return backend_compile(backend, computation, compile_options)
+  # Avoid import cycle between jax and jax.experimental
+  from jax.experimental.compilation_cache import compilation_cache as cc
+
+  if isinstance(computation, ir.Module):
+    module_name = computation.operation.name
+    computation = mlir.module_to_string(computation)
+  else:
+    module_name = computation.name()
+
+  # Persistent compilation cache only implemented on TPU.
+  # TODO(skye): add warning when initializing cache on unsupported default platform
+  if cc.is_initialized() and backend.platform == 'tpu':
+    cached_executable = cc.get_executable(computation, compile_options, backend)
+    if cached_executable is not None:
+      logging.info('Persistent compilation cache hit for %s.', module_name)
+      return cached_executable
+    else:
+      compiled = backend_compile(backend, computation, compile_options)
+      cc.put_executable(module_name, computation, compile_options, compiled,
+                        backend)
+      return compiled
+
+  if FLAGS.jax_dump_ir_to:
+    ir_str = (computation if isinstance(computation, str)
+              else computation.as_hlo_text())
+    _dump_ir_to_file(module_name, ir_str)
+  return backend_compile(backend, computation, compile_options)
 
 
 class XlaCompiledComputation:
@@ -560,7 +606,9 @@ class XlaCompiledComputation:
         num_partitions=1,
         device_assignment=(sticky_device.id,) if sticky_device else None)
     options.parameter_is_tupled_arguments = tuple_args
-    compiled = compile_or_get_cached(backend, xla_computation, options)
+    with log_elapsed_time(f"Finished XLA compilation of {name} "
+                          "in {elapsed_time} sec"):
+      compiled = compile_or_get_cached(backend, xla_computation, options)
     buffer_counts = (None if len(out_avals) == 1 else
                      [aval_to_num_buffers(aval) for aval in out_avals])
     execute = _execute_compiled if nreps == 1 else _execute_replicated
@@ -690,7 +738,7 @@ ad.deflinear2(device_put_p, lambda cotangent, _, **kwargs: [cotangent])
 masking.defvectorized(device_put_p)
 batching.defvectorized(device_put_p)
 
-def _device_put_lowering(ctx, avals_in, avals_out, x, *, device):
+def _device_put_lowering(ctx, x, *, device):
   return [x]
 
 
