@@ -71,7 +71,7 @@ def _check_prng_key(key):
           'Raw arrays as random keys to jax.random functions are deprecated. '
           'Assuming valid threefry2x32 key for now.',
           FutureWarning)
-    return prng.PRNGKeyArray(_get_default_prng_impl(), key), True
+    return prng.PRNGKeyArray(default_prng_impl(), key), True
   else:
     raise TypeError(f'unexpected PRNG key type {type(key)}')
 
@@ -94,7 +94,13 @@ PRNG_IMPLS = {
     'unsafe_rbg': prng.unsafe_rbg_prng_impl,
 }
 
-def _get_default_prng_impl():
+def default_prng_impl():
+  """Get the default PRNG implementation.
+
+  The default implementation is determined by ``config.jax_default_prng_impl``,
+  which specifies it by name. This function returns the corresponding
+  ``jax.prng.PRNGImpl`` instance.
+  """
   impl_name = config.jax_default_prng_impl
   assert impl_name in PRNG_IMPLS, impl_name
   return PRNG_IMPLS[impl_name]
@@ -117,13 +123,13 @@ def PRNGKey(seed: int) -> KeyArray:
     and ``fold_in``.
 
   """
-  impl = _get_default_prng_impl()
+  impl = default_prng_impl()
   key = prng.seed_with_impl(impl, seed)
   return _return_prng_keys(True, key)
 
 # TODO(frostig): remove once we always enable_custom_prng
 def _check_default_impl_with_no_custom_prng(impl, name):
-  default_impl = _get_default_prng_impl()
+  default_impl = default_prng_impl()
   default_name = config.jax_default_prng_impl
   if not config.jax_enable_custom_prng and default_impl is not impl:
     raise RuntimeError('jax_enable_custom_prng must be enabled in order '
@@ -1109,20 +1115,21 @@ def _poisson(key, lam, shape, dtype):
 
 def poisson(key: KeyArray,
             lam: RealArray,
-            shape: Sequence[int] = (),
+            shape: Optional[Sequence[int]] = None,
             dtype: DTypeLikeInt = dtypes.int_) -> jnp.ndarray:
   """Sample Poisson random values with given shape and integer dtype.
 
   Args:
     key: a PRNG key used as the random key.
-    lam: rate parameter (mean of the distribution), must be >= 0.
+    lam: rate parameter (mean of the distribution), must be >= 0. Must be broadcast-compatible with ``shape``
     shape: optional, a tuple of nonnegative integers representing the result
-      shape. Default ().
+      shape. Default (None) produces a result shape equal to ``lam.shape``.
     dtype: optional, a integer dtype for the returned values (default int64 if
       jax_enable_x64 is true, otherwise int32).
 
   Returns:
-    A random array with the specified shape and dtype.
+    A random array with the specified dtype and with shape given by ``shape`` if
+    ``shape is not None, or else by ``lam.shape``.
   """
   key, _ = _check_prng_key(key)
   if key.impl is not prng.threefry_prng_impl:
@@ -1130,9 +1137,11 @@ def poisson(key: KeyArray,
         '`poisson` is only implemented for the threefry2x32 RNG, '
         f'not {key.impl}')
   dtype = dtypes.canonicalize_dtype(dtype)
-  shape = core.canonicalize_shape(shape)
-  if np.shape(lam) != shape:
-    lam = jnp.broadcast_to(lam, shape)
+  if shape is not None:
+    shape = core.canonicalize_shape(shape)
+  else:
+    shape = np.shape(lam)
+  lam = jnp.broadcast_to(lam, shape)
   lam = lax.convert_element_type(lam, np.float32)
   return _poisson(key, lam, shape, dtype)
 
