@@ -41,7 +41,7 @@ from jax._src.lib import xla_bridge
 from jax._src import dispatch
 from jax.interpreters import mlir
 from jax.interpreters import xla
-from jax.experimental.maps import mesh
+from jax.experimental.maps import mesh, Mesh
 
 
 FLAGS = flags.FLAGS
@@ -92,6 +92,39 @@ def _dtype(x):
 def num_float_bits(dtype):
   return _dtypes.finfo(_dtypes.canonicalize_dtype(dtype)).bits
 
+def to_default_dtype(arr):
+  """Convert a value to an array with JAX's default dtype.
+
+  This is generally used for type conversions of values returned by numpy functions,
+  to make their dtypes take into account the state of the ``jax_enable_x64`` and
+  ``jax_default_dtype_bits`` flags.
+  """
+  arr = np.asarray(arr)
+  dtype = _dtypes._default_types.get(arr.dtype.kind)
+  return arr.astype(_dtypes.canonicalize_dtype(dtype)) if dtype else arr
+
+def with_jax_dtype_defaults(func, use_defaults=True):
+  """Return a version of a function with outputs that match JAX's default dtypes.
+
+  This is generally used to wrap numpy functions within tests, in order to make
+  their default output dtypes match those of corresponding JAX functions, taking
+  into account the state of the ``jax_enable_x64`` and ``jax_default_dtype_bits``
+  flags.
+
+  Args:
+    use_defaults : whether to convert any given output to the default dtype. May be
+      a single boolean, in which case it specifies the conversion for all outputs,
+      or may be a a pytree with the same structure as the function output.
+  """
+  @functools.wraps(func)
+  def wrapped(*args, **kwargs):
+    result = func(*args, **kwargs)
+    if isinstance(use_defaults, bool):
+      return tree_map(to_default_dtype, result) if use_defaults else result
+    else:
+      f = lambda arr, use_default: to_default_dtype(arr) if use_default else arr
+      return tree_map(f, result, use_defaults)
+  return wrapped
 
 def is_sequence(x):
   try:
@@ -976,6 +1009,9 @@ class JaxTestCase(parameterized.TestCase):
     ignore_space_re = re.compile(r'\s*\n\s*')
     expected_clean = re.sub(ignore_space_re, '\n', expected.strip())
     what_clean = re.sub(ignore_space_re, '\n', what.strip())
+    if what_clean != expected_clean:
+      # Print it so we can copy-and-paste it into the test
+      print(f"Found\n{what}\n")
     self.assertMultiLineEqual(expected_clean, what_clean,
                               msg="Found\n{}\nExpecting\n{}".format(what, expected))
 
@@ -1093,6 +1129,16 @@ def restore_spmd_lowering_flag():
   if old_spmd_lowering_flag is None: return
   config.update('experimental_xmap_spmd_lowering', old_spmd_lowering_flag)
 
+def create_global_mesh(mesh_shape, axis_names):
+  size = prod(mesh_shape)
+  if len(api.devices()) < size:
+    raise unittest.SkipTest(f"Test requires {size} global devices.")
+  devices = sorted(api.devices(), key=lambda d: d.id)
+  mesh_devices = np.array(devices[:size]).reshape(mesh_shape)
+  global_mesh = Mesh(mesh_devices, axis_names)
+  return global_mesh
+
+
 class _cached_property:
   null = object()
 
