@@ -16,6 +16,7 @@
 import collections
 from collections import namedtuple
 from contextlib import contextmanager
+import functools
 from functools import partial, partialmethod, total_ordering
 import gc
 import itertools as it
@@ -260,7 +261,6 @@ class Primitive:
   multiple_results = False  # set for multi-output primitives
   call_primitive = False    # set for call primitives processed in final style
   map_primitive = False     # set for map primitives processed in final style
-  _dispatch_on_params = False  # whether to include axis names from params in dispatch
 
   def __init__(self, name: str):
     self.name = name
@@ -268,14 +268,13 @@ class Primitive:
   def __repr__(self):
     return '{}'.format(self.name)
 
-
   def bind(self, *args, **params):
     assert (not config.jax_enable_checks or
             all(isinstance(arg, Tracer) or valid_jaxtype(arg) for arg in args)), args
-    top_trace = find_top_trace(
-        args, used_axis_names(self, params) if self._dispatch_on_params else None)
-    tracers = map(top_trace.full_raise, args)
-    out = top_trace.process_primitive(self, tracers, params)
+    return self.bind_with_trace(find_top_trace(args), args, params)
+
+  def bind_with_trace(self, trace, args, params):
+    out = trace.process_primitive(self, map(trace.full_raise, args), params)
     return map(full_lower, out) if self.multiple_results else full_lower(out)
 
   def def_impl(self, impl):
@@ -298,28 +297,12 @@ class Primitive:
     raise NotImplementedError("Abstract evaluation for '{}' not implemented"
                               .format(self.name))
 
+  def get_bind_params(self, params):
+    return [], params
+
 
 # -------------------- lifting --------------------
 
-# TODO(necula): this belongs next to pe.new_eqn_recipe, but is needed in
-# core.py. Plan to move all these utilities to jaxpr.py.
-def extract_call_jaxpr(
-  primitive: Primitive,
-  params: Dict[str, Any]) -> Tuple[Optional[Jaxpr], Dict[str, Any]]:
-  """Extract the call primitive subjaxpr from the params.
-
-  Returns the subjaxpr and the params without the "call_jaxpr" value. If this is
-  not a call primitive then returns (None, params).
-  """
-  if not (primitive.call_primitive or primitive.map_primitive):
-    return (None, params)
-  else:
-    assert "call_jaxpr" in params
-    new_params = dict(params)
-    del new_params["call_jaxpr"]
-    return (params["call_jaxpr"], new_params)
-
-
 # TODO(mattjj): replace this approach with a primitive-keyed table of rules
 def traverse_jaxpr_params(f, params):
   """Applies f to each jaxpr parameter and returns a tuple of returned values."""
@@ -329,26 +312,6 @@ def traverse_jaxpr_params(f, params):
           if type(p) in (Jaxpr, ClosedJaxpr)}
 
 
-def eval_jaxpr_eqn(eqn, in_vals):
-  """Evaluates the jaxpr equation with the provided input values."""
-  call_jaxpr, params = extract_call_jaxpr(eqn.primitive, eqn.params)
-  if call_jaxpr:
-    subfuns = [lu.wrap_init(partial(eval_jaxpr, call_jaxpr, ()))]
-  else:
-    subfuns = []
-  if eqn.primitive in initial_to_final_param_rules:
-    bind_params = initial_to_final_param_rules[eqn.primitive](params)
-  elif eqn.primitive.map_primitive:
-    out_axes_thunk = HashableFunction(lambda: params['out_axes'],
-                                      closure=params['out_axes'])
-    bind_params = dict(params, out_axes_thunk=out_axes_thunk)
-    del bind_params['out_axes']
-  else:
-    bind_params = params
-  with source_info_util.user_context(eqn.source_info.traceback):
-    return eqn.primitive.bind(*(subfuns + in_vals), **bind_params)
-
-
 def eval_jaxpr(jaxpr: Jaxpr, consts, *args):
   def read(v):
     if type(v) is Literal:
@@ -364,15 +327,15 @@ def eval_jaxpr(jaxpr: Jaxpr, consts, *args):
   map(write, jaxpr.constvars, consts)
   map(write, jaxpr.invars, args)
   for eqn in jaxpr.eqns:
-    ans = eval_jaxpr_eqn(eqn, map(read, eqn.invars))
+    subfuns, bind_params = eqn.primitive.get_bind_params(eqn.params)
+    with source_info_util.user_context(eqn.source_info.traceback):
+      ans = eqn.primitive.bind(*subfuns, *map(read, eqn.invars), **bind_params)
     if eqn.primitive.multiple_results:
       map(write, eqn.outvars, ans)
     else:
       write(eqn.outvars[0], ans)
   return map(read, jaxpr.outvars)
 
-initial_to_final_param_rules: Dict[Primitive, Callable] = {}
-
 
 # -------------------- tracing --------------------
 
@@ -852,9 +815,67 @@ def new_base_main(trace_type: Type[Trace]) -> Generator[MainTrace, None, None]:
         raise Exception(f'Leaked level {t()}. Leaked tracer(s): {leaked_tracers}.')
 
 @contextmanager
-def eval_context():
+def ensure_compile_time_eval():
+  """Context manager to ensure evaluation at trace/compile time (or error).
+
+  Some JAX APIs like ``jax.jit`` and ``jax.lax.scan`` involve staging, i.e.
+  delaying the evaluation of numerical expressions (like jax.numpy function
+  applications) so that instead of performing those computations eagerly while
+  evaluating the corresponding Python expressions, their computation is carried
+  out separately, e.g. after optimized compilation. But this delay can be
+  undesirable. For example, numerical values might be needed to evaluate Python
+  control flow and so their evaluation cannot be delayed. As another example, it
+  may be beneficial to ensure compile time evaluation (or "constant folding")
+  for performance reasons.
+
+  This context manager ensures that JAX computations are evaluated eagerly. If
+  eager evaluation is not possible, a ``ConcretizationError`` is raised.
+
+  Here's a contrived example::
+
+    import jax
+    import jax.numpy as jnp
+
+    @jax.jit
+    def f(x):
+      with jax.ensure_compile_time_eval():
+        y = jnp.sin(3.0)
+        z = jnp.sin(y)
+      if z > 0:  # the value of z is availble and can be used in control flow
+        return jnp.sin(x)
+      else:
+        return jnp.cos(x)
+
+  Here's a real-world example from https://github.com/google/jax/issues/3974::
+
+    import jax
+    import jax.numpy as jnp
+    from jax import random
+
+    @jax.jit
+    def jax_fn(x):
+      with jax.ensure_compile_time_eval():
+        y = random.randint(random.PRNGKey(0), (1000,1000), 0, 100)
+      y2 = y @ y
+      x2 = jnp.sum(y2) * x
+      return x2
+
+  A similar behavior can often be achieved simply by 'hoisting' the constant
+  expression out of the corresponding staging API::
+
+    y = random.randint(random.PRNGKey(0), (1000,1000), 0, 100)
+
+    @jax.jit
+    def jax_fn(x):
+      y2 = y @ y
+      x2 = jnp.sum(y2)*x
+      return x2
+
+  But in some cases it can be more convenient to use this context manager.
+  """
   with new_base_main(EvalTrace):
     yield
+eval_context = ensure_compile_time_eval  # alias, backward compatibility
 
 @contextmanager
 def new_sublevel() -> Generator[None, None, None]:
@@ -879,17 +900,14 @@ def full_lower(val):
   else:
     return val
 
-def find_top_trace(xs, axis_names=None) -> Trace:
-  top_main: Optional[MainTrace] = None
-  if axis_names:
-    top_main = max((axis_frame(a).main_trace for a in axis_names),
-                   default=None, key=lambda t: getattr(t, 'level', -1))
+def find_top_trace(xs) -> Trace:
   top_tracer = max((x for x in xs if isinstance(x, Tracer)),
-                   default=None, key=attrgetter('_trace.level'))
+                    default=None, key=attrgetter('_trace.level'))
   if top_tracer is not None:
     top_tracer._assert_live()
-    if top_tracer._trace.main.level > getattr(top_main, 'level', -1):
-      top_main = top_tracer._trace.main
+    top_main = top_tracer._trace.main
+  else:
+    top_main = None  # type: ignore
   dynamic = thread_local_state.trace_state.trace_stack.dynamic
   top_main = (dynamic if top_main is None or dynamic.level > top_main.level
               else top_main)
@@ -925,7 +943,7 @@ class AbstractValue:
     raise NotImplementedError("must override")
 
   def str_short(self, short_dtypes=False):
-    raise NotImplementedError("must override")
+    return str(self)
 
 class Bot(AbstractValue): pass
 
@@ -973,7 +991,7 @@ def check_valid_jaxtype(x):
 
 
 def concrete_aval(x):
-  for typ in type(x).mro():
+  for typ in type(x).__mro__:
     handler = pytype_aval_mappings.get(typ)
     if handler: return handler(x)
   if hasattr(x, '__jax_array__'):
@@ -1038,7 +1056,7 @@ def _short_dtype_name(dtype):
 
 class UnshapedArray(AbstractValue):
   __slots__ = ['dtype', 'weak_type']
-  array_abstraction_level = 2
+  array_abstraction_level = 3
 
   def __init__(self, dtype, weak_type=False):
     self.dtype = np.dtype(dtype)
@@ -1102,6 +1120,58 @@ class UnshapedArray(AbstractValue):
            "UnshapedArray instances to ever be produced.")
     raise TypeError(msg)
 
+
+# We have a convention of reusing AbsractValues as types, in particular reusing
+# ShapedArrays as types, even though we could make a distinction and use
+# abstract values during tracing only. This reuse becomes a bit more extreme
+# with DShapedArrays. A DShapedArray's shape attribute is a tuple which can
+# contain several different types: ints, other AbstractValues (specifically at
+# the input and output to pe.trace_to_jaxpr_dynamic), Tracers (while tracing),
+# or Vars (when used as jaxpr type annotations). We could reduce this
+# polymorphism if it seems cleaner, though it's kind of convenient!
+AxisSizeForTracing = Union[int, Tracer]
+AxisSizeForJaxprType = Union[int, Var]
+AxisSizeForJaxprTracingSpec = Union[int, AbstractValue]
+AxisSize = Union[AxisSizeForTracing, AxisSizeForJaxprType,
+                 AxisSizeForJaxprTracingSpec]
+
+class DShapedArray(UnshapedArray):
+  __slots__ = ['shape']
+  shape: Tuple[AxisSize, ...]  # see comment above
+  array_abstraction_level = 2
+
+  def __init__(self, shape, dtype, weak_type):
+    self.shape = shape
+    self.dtype = dtype
+    self.weak_type = weak_type
+
+  ndim = property(lambda self: len(self.shape))
+  size = property(lambda self: prod(self.shape))
+
+  def str_short(self, short_dtypes=False) -> str:
+    del short_dtypes  # ignored
+    shape = f'{",".join(str(d) for d in self.shape)}' if self.shape else ''
+    dtype = _short_dtype_name(self.dtype)
+    return f'{dtype}[{shape}]'
+  __str__ = __repr__ = str_short
+
+  def __eq__(self, other):
+    return (type(self) is type(other) and
+            self.dtype == other.dtype and self.shape == other.shape and
+            self.weak_type == other.weak_type)
+
+  def update(self, shape=None, dtype=None, weak_type=None):
+    if shape is None:
+      shape = self.shape
+    if dtype is None:
+      dtype = self.dtype
+    if weak_type is None:
+      weak_type = self.weak_type
+    return DShapedArray(shape, dtype, weak_type)
+
+del AxisSize, AxisSizeForTracing, AxisSizeForJaxprType, \
+    AxisSizeForJaxprTracingSpec
+
 class ShapedArray(UnshapedArray):
   __slots__ = ['shape', 'named_shape']
   array_abstraction_level = 1
@@ -1261,7 +1331,7 @@ pytype_aval_mappings[Token] = lambda _: abstract_token
 def raise_to_shaped(aval: AbstractValue, weak_type=None):
   if weak_type is None:
     weak_type = getattr(aval, 'weak_type', False)
-  for typ in type(aval).mro():
+  for typ in type(aval).__mro__:
     handler = raise_to_shaped_mappings.get(typ)
     if handler: return handler(aval, weak_type)
   raise TypeError(type(aval))
@@ -1593,25 +1663,32 @@ def as_named_shape(shape) -> NamedShape:
 
 # ------------------- Call -------------------
 
-def apply_todos(todos, outs):
-  todos_list = list(todos)
-  while todos_list:
-    outs = map(full_lower, todos_list.pop()(outs))
-  return outs
+class CallPrimitive(Primitive):
+  multiple_results = True
+  call_primitive = True
 
-class _IgnoreElemList(list):
-  """Compares equal to all other _ignore_elem_lists."""
-  def __hash__(self): return 0
-  def __eq__(self, other):
-    return type(other) is _IgnoreElemList
+  def bind(self, fun, *args, **params):
+    return call_bind(self, fun, *args, **params)
+
+  def get_bind_params(self, params):
+    new_params = dict(params)
+    subfun = lu.wrap_init(partial(eval_jaxpr, new_params.pop('call_jaxpr'), ()))
+    return [subfun], new_params
+
+def call_bind(primitive: CallPrimitive, fun, *args, **params):
+  top_trace = find_top_trace(args)
+  fun, env_trace_todo = process_env_traces_call(
+      fun, primitive, top_trace and top_trace.level, tuple(params.items()))
+  tracers = map(top_trace.full_raise, args)
+  outs = top_trace.process_call(primitive, fun, tracers, params)
+  return map(full_lower, apply_todos(env_trace_todo(), outs))
 
 @lu.transformation_with_aux
-def process_env_traces(primitive: Union['CallPrimitive', 'MapPrimitive'],
-                       level: int, params_tuple: tuple, out_axes_transforms, *args):
+def process_env_traces_call(primitive: CallPrimitive, level: int,
+                            params_tuple: tuple, *args):
   outs = yield args, {}
   params = dict(params_tuple)
   todo = []
-  assert not out_axes_transforms
   while True:
     tracers = [x for x in outs if isinstance(x, Tracer)
                and (level is None or x._trace.level > level)]
@@ -1621,51 +1698,17 @@ def process_env_traces(primitive: Union['CallPrimitive', 'MapPrimitive'],
       break
     trace = ans._trace.main.with_cur_sublevel()
     outs = map(trace.full_raise, outs)
-    outs, cur_todo = primitive.post_process(trace, outs, params)
-    if isinstance(primitive, MapPrimitive):
-      cur_todo, out_axes_transform = cur_todo
-      out_axes_transforms.append(out_axes_transform)
+    outs, cur_todo = trace.post_process_call(primitive, outs, params)
     todo.append(cur_todo)
   yield outs, tuple(todo)  # Ensure the aux output is immutable
 
-def call_bind(primitive: Union['CallPrimitive', 'MapPrimitive'],
-              fun, *args, **params):
-  out_axes_transforms = _IgnoreElemList()
-  if primitive.map_primitive:
-    out_axes_thunk = params['out_axes_thunk']
-    # The new thunk depends deterministically on the old thunk and the wrapped
-    # function. Any caching already has to include the wrapped function as part
-    # of the key, so we only use the previous thunk for equality checks.
-    @as_hashable_function(closure=out_axes_thunk)
-    def new_out_axes_thunk():
-      out_axes = out_axes_thunk()
-      for t in out_axes_transforms:
-        out_axes = t(out_axes)
-      return out_axes
-    params = dict(params, out_axes_thunk=new_out_axes_thunk)
-  params_tuple = tuple(params.items())
-  top_trace = find_top_trace(args)
-  fun, env_trace_todo = process_env_traces(
-      fun, primitive, top_trace and top_trace.level,
-      params_tuple, out_axes_transforms)
-  tracers = map(top_trace.full_raise, args)
-  outs = primitive.process(top_trace, fun, tracers, params)
-  return map(full_lower, apply_todos(env_trace_todo(), outs))
+def apply_todos(todos, outs):
+  todos_list = list(todos)
+  while todos_list:
+    outs = map(full_lower, todos_list.pop()(outs))
+  return outs
 
 
-class CallPrimitive(Primitive):
-  multiple_results = True
-  call_primitive = True
-
-  def bind(self, fun, *args, **params):
-    return call_bind(self, fun, *args, **params)
-
-  def process(self, trace, fun, tracers, params):
-    return trace.process_call(self, fun, tracers, params)
-
-  def post_process(self, trace, out_tracers, params):
-    return trace.post_process_call(self, out_tracers, params)
-
 def call_impl(f: lu.WrappedFun, *args, **params):
   del params  # params parameterize the call primitive, not the function
   with new_sublevel():
@@ -1678,8 +1721,99 @@ call_p.def_impl(call_impl)
 named_call_p: CallPrimitive = CallPrimitive('named_call')
 named_call_p.def_impl(call_impl)
 
+
+outfeed_primitives: Set[Primitive] = set()
+def jaxpr_uses_outfeed(jaxpr: Jaxpr) -> bool:
+  """Finds if there are outfeed primitives anywhere inside a Jaxpr."""
+  return any(primitive_uses_outfeed(eqn.primitive, eqn.params)
+             for eqn in jaxpr.eqns)
+
+def _param_uses_outfeed(param):
+  if type(param) is Jaxpr:
+    if jaxpr_uses_outfeed(param):
+      return True
+  elif type(param) is ClosedJaxpr:
+    if jaxpr_uses_outfeed(param.jaxpr):
+      return True
+  return False
+
+def primitive_uses_outfeed(prim: Primitive, params: Dict) -> bool:
+  if prim in outfeed_primitives:
+    return True
+  for param in params.values():
+    if isinstance(param, tuple):
+      if any(unsafe_map(_param_uses_outfeed, param)):
+        return True
+    elif _param_uses_outfeed(param):
+      return True
+  return False
+
 # ------------------- Map -------------------
 
+class MapPrimitive(Primitive):
+  multiple_results = True
+  map_primitive = True
+
+  def bind(self, fun, *args, **params):
+    assert len(params['in_axes']) == len(args)
+    return map_bind(self, fun, *args, **params)
+
+  def process(self, trace, fun, tracers, params):
+    return trace.process_map(self, fun, tracers, params)
+
+  def post_process(self, trace, out_tracers, params):
+    return trace.post_process_map(self, out_tracers, params)
+
+  def get_bind_params(self, params):
+    new_params = dict(params)
+    subfun = lu.wrap_init(partial(eval_jaxpr, new_params.pop('call_jaxpr'), ()))
+    axes = new_params.pop('out_axes')
+    new_params['out_axes_thunk'] = HashableFunction(lambda: axes, closure=axes)
+    return [subfun], new_params
+
+def map_bind(primitive: 'MapPrimitive', fun, *args, out_axes_thunk, **params):
+  # The new thunk depends deterministically on the old thunk and the wrapped
+  # function. Any caching already has to include the wrapped function as part
+  # of the key, so we only use the previous thunk for equality checks.
+  @as_hashable_function(closure=out_axes_thunk)
+  def new_out_axes_thunk():
+    out_axes = out_axes_thunk()
+    _, out_axes_transforms = todo_and_xforms()
+    for t in out_axes_transforms:
+      out_axes = t(out_axes)
+    return out_axes
+  params = dict(params, out_axes_thunk=new_out_axes_thunk)
+  params_tuple = tuple(params.items())
+  top_trace = find_top_trace(args)
+  fun, todo_and_xforms = process_env_traces_map(
+      fun, primitive, top_trace and top_trace.level, params_tuple)
+  tracers = map(top_trace.full_raise, args)
+  outs = primitive.process(top_trace, fun, tracers, params)
+  env_trace_todo, _ = todo_and_xforms()
+  return map(full_lower, apply_todos(env_trace_todo, outs))
+
+@lu.transformation_with_aux
+def process_env_traces_map(primitive: MapPrimitive, level: int,
+                           params_tuple: tuple, *args):
+  outs = yield args, {}
+  params = dict(params_tuple)
+  todo = []
+  out_axes_transforms = []
+  while True:
+    tracers = [x for x in outs if isinstance(x, Tracer)
+               and (level is None or x._trace.level > level)]
+    if tracers:
+      ans = max(tracers, key=lambda x: x._trace.level)
+    else:
+      break
+    trace = ans._trace.main.with_cur_sublevel()
+    outs = map(trace.full_raise, outs)
+    outs, (cur_todo, cur_xform) = primitive.post_process(trace, outs, params)
+    todo.append(cur_todo)
+    out_axes_transforms.append(cur_xform)
+  yield outs, (tuple(todo), tuple(out_axes_transforms))
+
+
 def mapped_aval(size: int, axis: int, aval: AbstractValue) -> AbstractValue:
   handler, _ = aval_mapping_handlers.get(type(aval), (None, None))
   if handler is not None:
@@ -1717,21 +1851,6 @@ aval_mapping_handlers: Dict[Type, AvalMapHandlerPair] = {
     ConcreteArray: (_map_shaped_array, _unmap_shaped_array),
 }
 
-
-class MapPrimitive(Primitive):
-  multiple_results = True
-  map_primitive = True
-
-  def bind(self, fun, *args, **params):
-    assert len(params['in_axes']) == len(args)
-    return call_bind(self, fun, *args, **params)
-
-  def process(self, trace, fun, tracers, params):
-    return trace.process_map(self, fun, tracers, params)
-
-  def post_process(self, trace, out_tracers, params):
-    return trace.post_process_map(self, out_tracers, params)
-
 @contextmanager
 def extend_axis_env(axis_name: AxisName, size: int, tag: Any):
   frame = AxisEnvFrame(axis_name, size, tag)
@@ -1889,7 +2008,14 @@ axis_substitution_rules: Dict[Primitive, Callable[[ParamDict, AxisSubst, bool],
 # participate in dispatch should subclass AxisPrimitive.
 
 class AxisPrimitive(Primitive):
-  _dispatch_on_params = True
+  def bind(self, *args, **params):
+    top_trace = find_top_trace(args)
+    axis_main = max((axis_frame(a).main_trace for a in used_axis_names(self, params)),
+                    default=None, key=lambda t: getattr(t, 'level', -1))
+    top_trace = (top_trace if not axis_main or axis_main.level < top_trace.level
+                 else axis_main.with_cur_sublevel())
+    return self.bind_with_trace(top_trace, args, params)
+
 
 # ------------------- Jaxpr checking -------------------
 
@@ -1925,10 +2051,6 @@ def typematch(aval1: AbstractValue, aval2: AbstractValue) -> bool:
 
 class JaxprTypeError(TypeError): pass
 
-def typecheck_assert(pred, msg):
-  if not pred:
-    raise JaxprTypeError(msg)
-
 custom_typechecks: Dict[Primitive, Callable] = {}
 
 def check_jaxpr(jaxpr: Jaxpr):
@@ -1942,34 +2064,48 @@ def check_jaxpr(jaxpr: Jaxpr):
   Raises `JaxprTypeError` if `jaxpr` is determined invalid. Returns `None`
   otherwise.
   """
+  @functools.lru_cache(maxsize=None)
+  def ctx_factory():
+    ctx = JaxprPpContext()
+    try: pp_jaxpr(jaxpr, ctx)  # side-effect on ctx, build variable names
+    except: pass
+    return ctx
+
   try:
-    _check_jaxpr(jaxpr, [v.aval for v in jaxpr.invars])
+    _check_jaxpr(ctx_factory, jaxpr, [v.aval for v in jaxpr.invars])
   except JaxprTypeError as e:
+    ctx = ctx_factory()
     if len(e.args) == 2:
-      msg, eqn_idx = e.args
-      jaxpr_str = str(pp_jaxpr_eqn_range(jaxpr, eqn_idx - 10, eqn_idx + 10,
-                                         JaxprPpContext()))
+      msg, eqnidx = e.args
+      jaxpr_str = str(pp_jaxpr_eqn_range(jaxpr, eqnidx - 10, eqnidx + 10, ctx))
     else:
       msg, = e.args
-      jaxpr_str = str(pp_jaxpr_eqn_range(jaxpr, 0, 20, JaxprPpContext()))
+      jaxpr_str = str(pp_jaxpr_eqn_range(jaxpr, 0, 20, ctx))
     msg = "\n\n".join([msg, "while checking jaxpr:", jaxpr_str])
     raise JaxprTypeError(msg) from None
 
-def _check_jaxpr(jaxpr: Jaxpr, in_avals: Sequence[AbstractValue]):
+def _check_jaxpr(ctx_factory: Callable[[], 'JaxprPpContext'], jaxpr: Jaxpr,
+                 in_avals: Sequence[AbstractValue]) -> None:
 
   def read(v: Atom) -> AbstractValue:
     if isinstance(v, Literal):
       return raise_to_shaped(get_aval(v.val))
     else:
-      typecheck_assert(v in env, f"Variable '{v}' not defined")
+      if v not in env:
+        ctx = ctx_factory()
+        raise JaxprTypeError(f"Variable '{pp_var(v, ctx)}' not defined")
       return env[v]
 
   def write(v: Var, a: AbstractValue) -> None:
-    typecheck_assert(v not in env, f"Variable '{v}' already bound")
+    if v in env:
+      ctx = ctx_factory()
+      raise JaxprTypeError(f"Variable '{pp_var(v, ctx)}' already bound")
     if not isinstance(v, DropVar):
-      typecheck_assert(typecompat(v.aval, a),
-                       f"Variable '{v}' inconsistently typed as {a}, "
-                       f"bound as {v.aval}")
+      if not typecompat(v.aval, a):
+        ctx = ctx_factory()
+        raise JaxprTypeError(
+            f"Variable '{pp_var(v, ctx)}' inconsistently typed as "
+            f"{pp_aval(a, ctx)}, bound as {pp_aval(v.aval, ctx)}")
       env[v] = a
 
   env : Dict[Var, AbstractValue] = {}
@@ -1982,24 +2118,24 @@ def _check_jaxpr(jaxpr: Jaxpr, in_avals: Sequence[AbstractValue]):
     prim = eqn.primitive
     try:
       in_avals = map(read, eqn.invars)
-      typecheck_assert(all(not isinstance(ina, ConcreteArray) for ina in in_avals),
-                       "Equation given ConcreteArray type inputs")
+      if any(isinstance(ina, ConcreteArray) for ina in in_avals):
+        raise JaxprTypeError("Equation given ConcreteArray type inputs")
       if prim in custom_typechecks:
         out_avals = custom_typechecks[prim](*in_avals, **eqn.params)
         if out_avals is None:
           out_avals = [v.aval for v in eqn.outvars]
       elif prim.call_primitive:
-        out_avals = check_call(prim, in_avals, eqn.params)
+        out_avals = check_call(ctx_factory, prim, in_avals, eqn.params)
       elif prim.map_primitive:
-        out_avals = check_map(prim, in_avals, eqn.params)
+        out_avals = check_map(ctx_factory, prim, in_avals, eqn.params)
       else:
         out_avals = check_eqn(prim, in_avals, eqn.params)
       map(write, eqn.outvars, out_avals)
     except JaxprTypeError as e:
+      ctx = ctx_factory()
       msg, = e.args
       src = source_info_util.summarize(eqn.source_info)
-      msg = "\n\n".join([msg, "in equation:",
-                        str(pp.nest(2, pp_eqn(eqn, JaxprPpContext()))),
+      msg = "\n\n".join([msg, "in equation:", str(pp.nest(2, pp_eqn(eqn, ctx))),
                          f"from source: {src}"])
       raise JaxprTypeError(msg, eqn_idx) from None
 
@@ -2014,57 +2150,58 @@ def check_eqn(prim, in_avals, params):
     out_avals = [out_avals]
   return out_avals
 
-def check_call(prim, in_avals, params):
-  typecheck_assert("call_jaxpr" in params,
-                   f"Call primitive {prim} missing 'call_jaxpr' parameter")
+def check_call(ctx_factory, prim, in_avals, params):
+  if "call_jaxpr" not in params:
+    raise JaxprTypeError(
+        f"Call primitive {prim} missing 'call_jaxpr' parameter")
   call_jaxpr = params["call_jaxpr"]
 
   # These checks also happen in recursive call, but give better errors here.
-  typecheck_assert(len(in_avals) == len(call_jaxpr.invars),
-                   f"Call primitive {prim} with {len(call_jaxpr.invars)} "
-                   f"operands cannot call jaxpr with {len(call_jaxpr.invars)} "
-                   f"inputs")
+  if len(in_avals) != len(call_jaxpr.invars):
+    raise JaxprTypeError(f"Call primitive {prim} with {len(call_jaxpr.invars)} "
+                         f"operands cannot call jaxpr with {len(call_jaxpr.invars)} "
+                         f"inputs")
   binder_avals = [v.aval for v in call_jaxpr.invars]
   for binder_aval, in_aval in zip(binder_avals, in_avals):
-    typecheck_assert(typecompat(binder_aval, in_aval),
-                     f"Call primitive {prim} passes operand {in_aval} "
-                     f"to jaxpr expecting {binder_aval}")
+    if not typecompat(binder_aval, in_aval):
+      raise JaxprTypeError(f"Call primitive {prim} passes operand {in_aval} "
+                           f"to jaxpr expecting {binder_aval}")
 
-  _check_jaxpr(call_jaxpr, in_avals)
+  _check_jaxpr(ctx_factory, call_jaxpr, in_avals)
 
   out_avals = [v.aval for v in call_jaxpr.outvars]
   return out_avals
 
-def check_map(prim, in_avals, params):
-  typecheck_assert("call_jaxpr" in params,
-                   f"Map primitive {prim} missing 'call_jaxpr' parameter")
+def check_map(ctx_factory, prim, in_avals, params):
+  if "call_jaxpr" not in params:
+    raise JaxprTypeError(f"Map primitive {prim} missing 'call_jaxpr' parameter")
   call_jaxpr = params["call_jaxpr"]
-  typecheck_assert("axis_size" in params,
-                   f"Map primitive {prim} missing 'axis_size' parameter")
+  if "axis_size" not in params:
+    raise JaxprTypeError(f"Map primitive {prim} missing 'axis_size' parameter")
   axis_size = params["axis_size"]
-  typecheck_assert("axis_name" in params,
-                   f"Map primitive {prim} missing 'axis_name' parameter")
+  if "axis_name" not in params:
+    raise JaxprTypeError(f"Map primitive {prim} missing 'axis_name' parameter")
   axis_name = params["axis_name"]
-  typecheck_assert("in_axes" in params,
-                   f"Map primitive {prim} missing 'in_axes' parameter")
+  if "in_axes" not in params:
+    raise JaxprTypeError(f"Map primitive {prim} missing 'in_axes' parameter")
   in_axes = params["in_axes"]
-  typecheck_assert("out_axes" in params,
-                   f"Map primitive {prim} missing 'out_axes' parameter")
+  if "out_axes" not in params:
+    raise JaxprTypeError(f"Map primitive {prim} missing 'out_axes' parameter")
   out_axes = params["out_axes"]
 
   binder_avals = [unmapped_aval(axis_size, axis_name, in_axis, v.aval)
                   if in_axis is not None else v.aval
                   for v, in_axis in zip(call_jaxpr.invars, in_axes)]
   for binder_aval, in_aval in zip(binder_avals, in_avals):
-    typecheck_assert(typecompat(binder_aval, in_aval),
-                     f"Call primitive {prim} passes operand {in_aval} "
-                     f"to jaxpr expecting {binder_aval}")
+    if not typecompat(binder_aval, in_aval):
+      raise JaxprTypeError(f"Call primitive {prim} passes operand {in_aval} "
+                           f"to jaxpr expecting {binder_aval}")
 
   mapped_avals = [mapped_aval(axis_size, in_axis, aval)
                   if in_axis is not None else aval
                   for aval, in_axis in zip(in_avals, in_axes)]
   with extend_axis_env(params['axis_name'], axis_size, None):
-    _check_jaxpr(call_jaxpr, mapped_avals)
+    _check_jaxpr(ctx_factory, call_jaxpr, mapped_avals)
 
   mapped_out_avals = [v.aval for v in call_jaxpr.outvars]
   out_avals = [unmapped_aval(axis_size, axis_name, out_axis, aval) if out_axis is not None else aval
@@ -2083,17 +2220,25 @@ class JaxprPpContext:
     self.var_ids = collections.defaultdict(it.count().__next__)
 
 
-def pp_var(v: Var, context: JaxprPpContext):
+def pp_var(v: Var, context: JaxprPpContext) -> str:
   if isinstance(v, (Literal, DropVar)): return str(v)
   return f"{_encode_digits_alphabetic(context.var_ids[v])}{v.suffix}"
 
+def pp_aval(a: AbstractValue, context: JaxprPpContext) -> str:
+  if isinstance(a, DShapedArray):
+    shape = [pp_var(d, context) if type(d) is Var else str(d) for d in a.shape]
+    dtype = _short_dtype_name(a.dtype)
+    return f'{dtype}[{",".join(shape)}]'
+  else:
+    return a.str_short(short_dtypes=True)
+
 def pp_vars(vs: Sequence[Any], context: JaxprPpContext,
             *, separator="", print_shapes: bool = False) -> pp.Doc:
   if print_shapes:
     return pp.nest(2, pp.group(
       pp.join(pp.text(separator) + pp.group(pp.brk()), [
         pp.text(pp_var(v, context)) +
-        pp.dim(pp.text(":" + v.aval.str_short(short_dtypes=True)))
+        pp.dim(pp.text(":" + pp_aval(v.aval, context)))
         for v in vs
       ])
     ))
@@ -2208,3 +2353,10 @@ def pp_jaxpr_eqn_range(jaxpr: Jaxpr, lo: int, hi: int, context: JaxprPpContext,
         pps.append(pp.text('...'))
     return pp.join(pp.brk("; "), pps)
   return pp_jaxpr_skeleton(jaxpr, eqns_fn, context, print_shapes=print_shapes)
+
+
+# TODO(mattjj,frostig): remove these stubs, which are a temporary hack for
+# google-internal type checking
+extract_call_jaxpr: Callable
+eval_jaxpr_eqn: Callable
+initial_to_final_param_rules: Dict
