@@ -14,6 +14,7 @@
 """GlobalDeviceArray serialization and deserialization."""
 
 import asyncio
+import re
 from typing import Callable
 
 import jax
@@ -24,21 +25,26 @@ import numpy as np
 import tensorstore as ts
 
 
-async def create_async_gsda_from_callback(
+async def create_async_gda_from_callback(
     global_shape: gda.Shape,
     global_mesh: Mesh,
     mesh_axes: gda.MeshAxes,
     data_callback: Callable[[gda.Index], asyncio.Future],
 ):
-  indices = gda.get_shard_indices(global_shape, global_mesh, mesh_axes)
-  future_arrays = [data_callback(indices[d]) for d in global_mesh.local_devices]
+  global_idx_rid = gda.get_shard_indices_replica_ids(
+      global_shape, global_mesh, mesh_axes)
+  local_devices = global_mesh.local_devices
+  future_arrays = [data_callback(global_idx_rid[d][0])
+                   for d in local_devices]
   # Pause here and come back to `from_async_callback()` when future_arrays are
   # ready. device_put cannot happen with future_arrays.
   local_arrays = await asyncio.gather(*future_arrays)
 
   dbs = [jax.device_put(array, device)
-         for array, device in zip(local_arrays, global_mesh.local_devices)]
-  return gda.GlobalDeviceArray(global_shape, global_mesh, mesh_axes, dbs)
+         for array, device in zip(local_arrays, local_devices)]
+  local_idx_rid = dict((d, global_idx_rid[d]) for d in local_devices)
+  return gda.GlobalDeviceArray(global_shape, global_mesh, mesh_axes, dbs,
+                               gda._GdaFastPathArgs(local_idx_rid, local_devices))
 
 
 def _get_metadata(gda):
@@ -58,13 +64,20 @@ def _get_metadata(gda):
   }
 
 
-def get_tensorstore_spec(ckpt_path):
+def get_tensorstore_spec(ckpt_path: str):
   spec = {'driver': 'zarr', 'kvstore': {}}
-  # TODO(yashkatariya): Add GCS kvstore too.
-  spec['kvstore'] = {
-      'driver': 'file',
-      'path': ckpt_path,
-  }
+
+  if ckpt_path.startswith('gs://'):
+    m = re.fullmatch('^gs://([^/]*)/(.*)$', ckpt_path, re.DOTALL)
+    if m is None:
+      raise ValueError('The ckpt_path should contain the bucket name and the '
+                       f'file path inside the bucket. Got: {ckpt_path}')
+    gcs_bucket = m.group(1)
+    path_without_bucket = m.group(2)
+    spec['kvstore'] = {'driver': 'gcs', 'bucket': gcs_bucket,
+                       'path': path_without_bucket}
+  else:
+    spec['kvstore'] = {'driver': 'file', 'path': ckpt_path}
   return spec
 
 
@@ -97,4 +110,4 @@ async def async_deserialize(ckpt_path, mesh, mesh_axes, tensorstore_spec):
   async def cb(index):
     return await t[index].read()
 
-  return await create_async_gsda_from_callback(t.shape, mesh, mesh_axes, cb)
+  return await create_async_gda_from_callback(t.shape, mesh, mesh_axes, cb)
