@@ -15,12 +15,14 @@
 
 Specific JAX primitive conversion tests are in primitives_test."""
 
+import unittest
 from typing import Dict, Tuple
 
 from absl.testing import absltest
 from absl.testing import parameterized
 
 import jax
+from jax import ad_checkpoint
 from jax import dtypes
 from jax import lax
 from jax import numpy as jnp
@@ -136,9 +138,9 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     mk_sharded = lambda f: jax.pmap(lambda x: x)(f([n]))
     f_tf = tf.function(lambda x: x)
     self.assertAllClose(f_tf(mk_sharded(jnp.zeros)).numpy(),
-                        np.zeros([n]))
+                        jnp.zeros([n]))
     self.assertAllClose(f_tf(mk_sharded(jnp.ones)).numpy(),
-                        np.ones([n]))
+                        jnp.ones([n]))
 
   @jtu.skip_on_devices("gpu")
   def test_bfloat16_passed_by_tf(self):
@@ -635,25 +637,33 @@ class Jax2TfTest(tf_test_util.JaxToTfTestCase):
     self.TransformConvertAndCompare(f, arg, "grad")
     self.TransformConvertAndCompare(f, arg, "grad_vmap")
 
-  def test_remat1(self):
-    @jax.remat
+  @parameterized.named_parameters(jtu.cases_from_list(
+    dict(testcase_name=f"_{flavor}", flavor=flavor)
+    for flavor in ["old", "new"]))
+  def test_remat(self, flavor="old"):
     def f(x1):
       x2 = jnp.sin(x1)
       x3 = jnp.sin(x2)
       x4 = jnp.sin(x3)
-      return jnp.sum(x4)
+      return x4
+    remat_f = jax.remat(f) if flavor == "old" else ad_checkpoint.checkpoint(f)
 
     # The computation of grad_f computes "sin" 5 times, 3 for the forward pass
     # and then to rematerialize "x2" and "x3" in the backward pass.
-    arg = np.arange(3.)
-    self.TransformConvertAndCompare(f, arg, "grad")
-    # TODO: check that the TF code also computes "sin" 5 times
+    arg = np.array(3.)
+    # Check that we have a Sin under a conditional
+    f_tf = tf.function(jax2tf.convert(jax.grad(remat_f)), autograph=False)
+    f_tf_graph = f_tf.get_concrete_function(arg).graph.as_graph_def()
+    if flavor == "old":
+      raise unittest.SkipTest("TODO: CSE widget not yet implemented for old-style remat")
+    self.assertRegex(str(f_tf_graph),
+                     r'remat_checkpoint_/switch_case/indexed_case/Sin')
 
   def test_remat_free_var(self):
     def f(x):
       y = 2 * x
 
-      @jax.remat
+      @ad_checkpoint.checkpoint
       def g():
         return y
 
