@@ -707,7 +707,9 @@ def make_xmap_callable(fun: lu.WrappedFun,
   mapped_in_avals = [_delete_aval_axes(aval, in_axes, global_axis_sizes)
                      for aval, in_axes in zip(in_avals, in_axes)]
   with core.extend_axis_env_nd(global_axis_sizes.items()):
-    jaxpr, out_avals, consts = pe.trace_to_jaxpr_final(fun, mapped_in_avals)
+    with dispatch.log_elapsed_time(f"Finished tracing + transforming {fun.__name__} "
+                                   "for xmap in {elapsed_time} sec"):
+      jaxpr, out_avals, consts = pe.trace_to_jaxpr_final(fun, mapped_in_avals)
   out_axes = out_axes_thunk()
   _check_out_avals_vs_out_axes(out_avals, out_axes, global_axis_sizes)
   # NOTE: We don't use avals and all params, so only pass in the relevant parts (too lazy...)
@@ -848,15 +850,15 @@ class EvaluationPlan(NamedTuple):
 # -------- xmap primitive and its transforms --------
 
 # xmap has a different set of parameters than pmap, so we make it its own primitive type
-class XMapPrimitive(core.MapPrimitive):  # Not really a map, but it gives us a few good defaults
+class XMapPrimitive(core.MapPrimitive):
   def __init__(self):
     super().__init__('xmap')
     self.def_impl(xmap_impl)
     self.def_custom_bind(self.bind)
 
-  def bind(self, fun, *args, **params):
-    assert len(params['in_axes']) == len(args), (params['in_axes'], args)
-    return core.call_bind(self, fun, *args, **params)  # type: ignore
+  def bind(self, fun, *args, in_axes, **params):
+    assert len(in_axes) == len(args), (in_axes, args)
+    return core.map_bind(self, fun, *args, in_axes=in_axes, **params)
 
   def process(self, trace, fun, tracers, params):
     return trace.process_xmap(self, fun, tracers, params)
@@ -864,6 +866,19 @@ class XMapPrimitive(core.MapPrimitive):  # Not really a map, but it gives us a f
   def post_process(self, trace, out_tracers, params):
     raise NotImplementedError
 
+  def get_bind_params(self, params):
+    new_params = dict(params)
+    subfun = lu.wrap_init(partial(core.eval_jaxpr, new_params.pop('call_jaxpr'), ()))
+    axes = new_params.pop('out_axes')
+    new_params['out_axes_thunk'] = HashableFunction(lambda: axes, closure=axes)
+    spmd_axes = new_params.pop('spmd_out_axes')
+    if spmd_axes is not None:
+      new_params['spmd_out_axes_thunk'] = \
+          HashableFunction(lambda: spmd_axes, closure=spmd_axes)
+    else:
+      new_params['spmd_out_axes_thunk'] = None
+    return [subfun], new_params
+
 xmap_p = XMapPrimitive()
 core.EvalTrace.process_xmap = core.EvalTrace.process_call  # type: ignore
 def _process_xmap_default(self, call_primitive, f, tracers, params):
@@ -936,14 +951,15 @@ def _typecheck_xmap(
   binder_in_avals = [_insert_aval_axes(v.aval, a_in_axes, local_axis_sizes)
                      for v, a_in_axes in zip(call_jaxpr.invars, in_axes)]
   for binder_in_aval, in_aval in zip(binder_in_avals, in_avals):
-    core.typecheck_assert(
-        core.typecompat(binder_in_aval, in_aval),
+    if not core.typecompat(binder_in_aval, in_aval):
+      raise core.JaxprTypeError(
         f"xmap passes operand {in_aval} to jaxpr expecting {binder_in_aval}")
 
   mapped_in_avals = [_delete_aval_axes(a, a_in_axes, global_axis_sizes)
                      for a, a_in_axes in zip(in_avals, in_axes)]
   with core.extend_axis_env_nd(global_axis_sizes.items()):
-    core._check_jaxpr(call_jaxpr, mapped_in_avals)
+    core._check_jaxpr(lambda: core.JaxprPpContext(), call_jaxpr,
+                      mapped_in_avals)
 
   mapped_out_avals = [v.aval for v in call_jaxpr.outvars]
   out_avals = [_insert_aval_axes(a, a_out_axes, local_axis_sizes)
@@ -1276,22 +1292,6 @@ batching.BatchTrace.process_xmap = partialmethod(_batch_trace_process_xmap, Fals
 pxla.SPMDBatchTrace.process_xmap = partialmethod(_batch_trace_process_xmap, True)  # type: ignore
 
 
-def _xmap_initial_to_final_params(params):
-  out_axes_thunk = HashableFunction(lambda: params['out_axes'],
-                                    closure=params['out_axes'])
-  if params['spmd_out_axes'] is not None:
-    spmd_out_axes_thunk = HashableFunction(lambda: params['spmd_out_axes'],
-                                           closure=params['spmd_out_axes'])
-  else:
-    spmd_out_axes_thunk = None
-  bind_params = dict(params,
-                     out_axes_thunk=out_axes_thunk,
-                     spmd_out_axes_thunk=spmd_out_axes_thunk)
-  del bind_params['out_axes']
-  del bind_params['spmd_out_axes']
-  return bind_params
-core.initial_to_final_param_rules[xmap_p] = _xmap_initial_to_final_params
-
 # -------- nested xmap handling --------
 
 def _xmap_lowering_rule(*args, **kwargs):
@@ -1301,14 +1301,14 @@ def _xmap_lowering_rule(*args, **kwargs):
     return _xmap_lowering_rule_replica(*args, **kwargs)
 mlir.register_lowering(xmap_p, _xmap_lowering_rule)
 
-def _xmap_lowering_rule_replica(ctx, avals_in, avals_out, *in_nodes,
-                                   call_jaxpr, name,
-                                   in_axes, out_axes, donated_invars,
-                                   global_axis_sizes,
-                                   spmd_in_axes, spmd_out_axes,
-                                   positional_semantics,
-                                   axis_resources, resource_env, backend):
-  xla.check_backend_matches(backend, ctx.platform)
+def _xmap_lowering_rule_replica(ctx, *in_nodes,
+                                call_jaxpr, name,
+                                in_axes, out_axes, donated_invars,
+                                global_axis_sizes,
+                                spmd_in_axes, spmd_out_axes,
+                                positional_semantics,
+                                axis_resources, resource_env, backend):
+  xla.check_backend_matches(backend, ctx.module_context.platform)
   # The only way for any of those two assertions to be violated is when xmap
   # is using the SPMD lowering, but then this rule shouldn't even trigger.
   assert positional_semantics == _PositionalSemantics.LOCAL
@@ -1344,30 +1344,36 @@ def _xmap_lowering_rule_replica(ctx, avals_in, avals_out, *in_nodes,
   assert not consts
 
   tiled_ins = (
-      mlir.lower_fun(
-          partial(_tile, in_axes=arg_in_axes, axis_sizes=local_mesh_shape),
-          multiple_results=False)(ctx, [aval], None, in_node)[0]
-      if v.aval is not core.abstract_unit else in_node
-      for v, aval, in_node, arg_in_axes
-      in zip(call_jaxpr.invars, avals_in, in_nodes, mesh_in_axes))
+    mlir.lower_fun(partial(_tile, in_axes=arg_in_axes,
+                           axis_sizes=local_mesh_shape),
+                   multiple_results=False)(
+          mlir.LoweringRuleContext(module_context=ctx.module_context,
+                                   primitive=None,
+                                   avals_in=[aval], avals_out=None),
+          in_node)[0]
+    if v.aval is not core.abstract_unit else in_node
+    for v, aval, in_node, arg_in_axes
+    in zip(call_jaxpr.invars, ctx.avals_in, in_nodes, mesh_in_axes))
 
   # NOTE: We don't extend the resource env with the mesh shape, because those
   #       resources are already in scope! It's the outermost xmap that introduces
   #       them!
   # We in-line here rather than generating a Call HLO as in the xla_call
   # translation rule just because the extra tuple stuff is a pain.
-  sub_ctx = ctx.replace(
-      name_stack=xla.extend_name_stack(ctx.name_stack,
+  sub_ctx = ctx.module_context.replace(
+      name_stack=xla.extend_name_stack(ctx.module_context.name_stack,
                                        xla.wrap_name(name, 'xmap')))
   tiled_outs = mlir.jaxpr_subcomp(sub_ctx, vectorized_jaxpr, (), *tiled_ins)
 
   outs = [
       mlir.lower_fun(
           partial(_untile, out_axes=ans_out_axes, axis_sizes=local_mesh_shape,
-                  platform=ctx.platform),
+                  platform=ctx.module_context.platform),
           multiple_results=False)(
-          ctx, [vectorized_outvar.aval], None, tiled_out
-      )[0]
+              mlir.LoweringRuleContext(module_context=ctx.module_context,
+                                       primitive=None,
+                                       avals_in=[vectorized_outvar.aval],
+                                       avals_out=None), tiled_out)[0]
       if v.aval is not core.abstract_unit else tiled_out
       for v, vectorized_outvar, tiled_out, ans_out_axes
       in zip(call_jaxpr.outvars, vectorized_jaxpr.outvars, tiled_outs,
@@ -1375,12 +1381,12 @@ def _xmap_lowering_rule_replica(ctx, avals_in, avals_out, *in_nodes,
   return outs
 
 
-def _xmap_lowering_rule_spmd(ctx, avals_in, avals_out, *global_in_nodes,
+def _xmap_lowering_rule_spmd(ctx, *global_in_nodes,
                              call_jaxpr, name, in_axes, out_axes,
                              donated_invars, global_axis_sizes, spmd_in_axes,
                              spmd_out_axes, positional_semantics,
                              axis_resources, resource_env, backend):
-  xla.check_backend_matches(backend, ctx.platform)
+  xla.check_backend_matches(backend, ctx.module_context.platform)
   plan = EvaluationPlan.from_axis_resources(axis_resources, resource_env, global_axis_sizes)
 
   resource_call_jaxpr = plan.subst_axes_with_resources(call_jaxpr)
@@ -1407,7 +1413,7 @@ def _xmap_lowering_rule_spmd(ctx, avals_in, avals_out, *global_in_nodes,
   # NOTE: We don't extend the resource env with the mesh shape, because those
   #       resources are already in scope! It's the outermost xmap that introduces
   #       them!
-  global_in_avals = avals_in
+  global_in_avals = ctx.avals_in
   vectorized_jaxpr, global_out_avals, consts = pe.trace_to_jaxpr_dynamic(f, global_in_avals)
   assert not consts
 
@@ -1420,8 +1426,8 @@ def _xmap_lowering_rule_spmd(ctx, avals_in, avals_out, *global_in_nodes,
 
   # We in-line here rather than generating a Call HLO as in the xla_call
   # translation rule just because the extra tuple stuff is a pain.
-  sub_ctx = ctx.replace(
-      name_stack=xla.extend_name_stack(ctx.name_stack,
+  sub_ctx = ctx.module_context.replace(
+      name_stack=xla.extend_name_stack(ctx.module_context.name_stack,
                                        xla.wrap_name(name, 'xmap')))
   global_out_nodes = mlir.jaxpr_subcomp(sub_ctx, vectorized_jaxpr, (),
                                         *sharded_global_in_nodes)
@@ -1603,17 +1609,32 @@ def _xmap_translation_rule_spmd(ctx, avals_in, avals_out, *global_in_nodes,
   # NOTE: We don't extend the resource env with the mesh shape, because those
   #       resources are already in scope! It's the outermost xmap that introduces
   #       them!
-  global_in_avals = [core.ShapedArray(xla_type.dimensions(), xla_type.numpy_dtype())
-                     for in_node in global_in_nodes
-                     for xla_type in (ctx.builder.get_shape(in_node),)]
-  vectorized_jaxpr, global_out_avals, consts = pe.trace_to_jaxpr_dynamic(f, global_in_avals)
+  global_in_avals = [
+      core.ShapedArray(xla_type.dimensions(), xla_type.numpy_dtype())
+      for in_node in global_in_nodes
+      for xla_type in (ctx.builder.get_shape(in_node),)
+  ]
+  vectorized_jaxpr, global_out_avals, consts = pe.trace_to_jaxpr_dynamic(
+      f, global_in_avals)
   assert not consts
 
   global_sharding_spec = pxla.mesh_sharding_specs(mesh.shape, mesh.axis_names)
+
+  def set_sharding(node, aval, aval_axes):
+    sharding_proto = global_sharding_spec(aval, aval_axes).sharding_proto()
+    if not config.experimental_xmap_ensure_fixed_sharding:
+      # Do not specify sharding on other dimensions.
+      unspecified_dims = set(range(aval.ndim))
+      for axis in set(aval_axes.values()):
+        unspecified_dims.remove(axis)
+      return xla.set_sharding_proto(ctx.builder, node, sharding_proto,
+                                    unspecified_dims)
+    else:
+      return xla.set_sharding_proto(ctx.builder, node, sharding_proto)
+
   sharded_global_in_nodes = [
-    xla.set_sharding_proto(ctx.builder, node, global_sharding_spec(aval, aval_axes).sharding_proto())
-    if aval_axes else node
-    for node, aval, aval_axes in zip(global_in_nodes, global_in_avals, mesh_in_axes)
+      set_sharding(node, aval, aval_axes) if aval_axes else node for node, aval,
+      aval_axes in zip(global_in_nodes, global_in_avals, mesh_in_axes)
   ]
 
   # We in-line here rather than generating a Call HLO as in the xla_call
@@ -1625,9 +1646,8 @@ def _xmap_translation_rule_spmd(ctx, avals_in, avals_out, *global_in_nodes,
                                        *sharded_global_in_nodes)
 
   sharded_global_out_nodes = [
-    xla.set_sharding_proto(ctx.builder, node, global_sharding_spec(aval, aval_axes).sharding_proto())
-    if aval_axes else node
-    for node, aval, aval_axes in zip(global_out_nodes, global_out_avals, mesh_out_axes)
+      set_sharding(node, aval, aval_axes) if aval_axes else node for node, aval,
+      aval_axes in zip(global_out_nodes, global_out_avals, mesh_out_axes)
   ]
 
   return sharded_global_out_nodes
