@@ -25,6 +25,7 @@ from jax.experimental.global_device_array import GlobalDeviceArray as GDA
 from jax import core
 from jax import linear_util as lu
 from jax._src.api import _check_callable, _check_arg, Lowered
+from jax._src import dispatch
 from jax._src import source_info_util
 from jax._src.api_util import (argnums_partial_except, flatten_axes,
                                flatten_fun_nokwargs, _ensure_index_tuple,
@@ -45,9 +46,16 @@ from jax._src.util import (extend_name_stack, HashableFunction, safe_zip,
                          split_list, cache, tuple_insert)
 xops = xc._xla.ops
 
-class _FromGsdaSingleton:
+class _FromGdaSingleton:
   pass
-FROM_GDA = _FromGsdaSingleton()
+FROM_GDA = _FromGdaSingleton()
+
+def _is_from_gda(x):
+  # It's occasionally possible to end up with two FROM_GDA singletons (e.g. if
+  # pickling in_axis_resources and sending to other processes). Make sure this
+  # doesn't cause an error to avoid user confusion.
+  return isinstance(x, type(FROM_GDA))
+
 
 # TODO(yashkatariya): Add pjit microbenchmarks.
 def pjit(fun: Callable,
@@ -95,9 +103,14 @@ def pjit(fun: Callable,
     of ``pjit`` every process only "sees" its local piece of the input and output,
     corresponding to its local sub-mesh.
 
-    The SPMD model requires that the same multi-process ``pjit``'d functions must
-    be run in the same order on all processes, but they can be interspersed with
-    arbitrary operations running in a single process.
+    This means that each process's participating local devices must form a
+    _contiguous_ local sub-mesh within the full global mesh. A contiguous
+    sub-mesh is one where all of its devices are adjacent within the global
+    mesh, and form a rectangular prism.
+
+    The SPMD model also requires that the same multi-process ``pjit``'d
+    functions must be run in the same order on all processes, but they can be
+    interspersed with arbitrary operations running in a single process.
 
   Args:
     fun: Function to be compiled. Should be a pure function, as side-effects may
@@ -219,7 +232,7 @@ def pjit(fun: Callable,
     # TODO(yashkatariya): This is a hack. This should go away when avals have
     # is_global attribute.
     in_positional_semantics = tuple(
-        maps._PositionalSemantics.GLOBAL if type(a) is GDA else maps
+        maps._PositionalSemantics.GLOBAL if isinstance(a, GDA) else maps
         ._positional_semantics for a in args_flat)
     out_positional_semantics = maps._positional_semantics
     jaxpr, in_axis_resources_flat, out_axis_resources_flat = _pjit_jaxpr(
@@ -229,8 +242,8 @@ def pjit(fun: Callable,
         hashable_pytree(out_axis_resources),
         in_positional_semantics, out_positional_semantics,
         tuple(isinstance(a, GDA) for a in args_flat))
-    in_axis_resources_flat = tree_map(_canonicalize_spec, in_axis_resources_flat,
-                                      tuple(args_flat))
+    in_axis_resources_flat = tree_map(_maybe_replace_from_gda_with_pspec,
+                                      in_axis_resources_flat, tuple(args_flat))
     params = dict(
         jaxpr=jaxpr,
         in_axis_resources=in_axis_resources_flat,
@@ -291,28 +304,39 @@ def _pjit_jaxpr(fun, mesh, local_in_avals,
   in_axis_resources_flat = flatten_axis_resources(
         "pjit in_axis_resources", in_tree,
         in_axis_resources_thunk(), tupled_args=True)
+  canonicalized_in_axis_resources_flat = tree_map(_create_cpspec, in_axis_resources_flat)
   # This check should be above local_to_global call below otherwise if
   # `FROM_GDA` is passed to any input other than GDA, a ugly error message
   # will be raised because get_array_mapping (in local_to_global) of a
   # FROM_GDA cannot happen.
   tree_map(_check_resources_mismatch, in_axis_resources_flat, is_gda)
-  # If all inputs are GDAs, then the avals are global and the mesh should also
-  # be global. This split is because non-contiguous mesh can only be used if all
-  # inputs are GDAs.
-  if all(is_gda):
+  # If all inputs are either GDAs or fully replicated, then the avals are
+  # global and the mesh should also be global. This split is because
+  # non-contiguous mesh can only be used if all inputs are either GDAs or fully
+  # replicated.
+  # Use canonicalized in_axis_resources here because we want to treat P(None)
+  # and None (for example) as equivalent.
+  if all(((not _is_from_gda(p) and p.partitions == ()) or ig)
+         for p, ig in safe_zip(canonicalized_in_axis_resources_flat, is_gda)):
+    # Shapes should be checked against non canonicalized in_axis_resources.
+    # For example, partitions of () and ((),) are not equivalent, since the
+    # first one is a valid spec for a scalar value, while the second is not!
     _check_shapes_against_resources(
         "pjit arguments", mesh.is_multi_process, mesh.shape, local_in_avals,
         in_axis_resources_flat)
   else:
     _check_shapes_against_resources("pjit arguments", False, mesh.local_mesh.shape,
                                     local_in_avals, in_axis_resources_flat)
+
   global_in_avals = local_to_global(in_positional_semantics, mesh,
-                                    local_in_avals, in_axis_resources_flat)
+                                    local_in_avals, canonicalized_in_axis_resources_flat)
 
   prev_positional = maps._positional_semantics
   try:
     maps._positional_semantics = maps._PositionalSemantics.GLOBAL
-    jaxpr, global_out_avals, consts = pe.trace_to_jaxpr_dynamic(fun, global_in_avals)
+    with dispatch.log_elapsed_time(f"Finished tracing + transforming {fun.__name__} "
+                                   "for pjit in {elapsed_time} sec"):
+      jaxpr, global_out_avals, consts = pe.trace_to_jaxpr_dynamic(fun, global_in_avals)
   finally:
     maps._positional_semantics = prev_positional
   jaxpr = core.ClosedJaxpr(jaxpr, consts)
@@ -322,8 +346,10 @@ def _pjit_jaxpr(fun, mesh, local_in_avals,
         out_axis_resources_thunk(), tupled_args=False)
   _check_shapes_against_resources("pjit outputs", mesh.is_multi_process, mesh.shape,
                                   global_out_avals, out_axis_resources_flat)
+  canonicalized_out_axis_resources_flat = tree_map(_create_cpspec, out_axis_resources_flat)
   # lu.cache needs to be able to create weakrefs to outputs, so we can't return a plain tuple
-  return _ListWithW([jaxpr, in_axis_resources_flat, out_axis_resources_flat])
+  return _ListWithW([jaxpr, canonicalized_in_axis_resources_flat,
+                     canonicalized_out_axis_resources_flat])
 
 
 class SpecSync(IntEnum):
@@ -337,11 +363,11 @@ class SpecSync(IntEnum):
   IN_SYNC = 2  # Entirely in sync
 
 class ParsedPartitionSpec:
-  __slots__ = ('partitions', 'unsafe_user_spec', 'sync')
+  __slots__ = ('unsafe_user_spec', 'partitions', 'sync')
 
   def __init__(self, user_spec, partitions, sync=SpecSync.IN_SYNC):
-    self.partitions = tuple(partitions)
     self.unsafe_user_spec = user_spec
+    self.partitions = tuple(partitions)
     self.sync = sync
 
   @property
@@ -363,7 +389,7 @@ class ParsedPartitionSpec:
     return ParsedPartitionSpec(self.unsafe_user_spec, new_partitions, sync=new_sync)
 
   @classmethod
-  def from_user_input(cls, entry, arg_name):
+  def from_user_input(cls, entry, arg_name, allow_unconstrained_dims=False):
     if entry is None:
       return cls(entry, ())
     if not isinstance(entry, PartitionSpec):
@@ -375,6 +401,10 @@ class ParsedPartitionSpec:
         axis_spec = ()
       elif isinstance(axis_spec, (list, tuple)):
         axis_spec = tuple(axis_spec)
+      elif axis_spec == PartitionSpec.UNCONSTRAINED:
+        if not allow_unconstrained_dims:
+          raise ValueError(f"Unconstrained dims are not allowed: {entry}")
+        axis_spec = None
       else:
         axis_spec = (axis_spec,)
       axis_specs.append(axis_spec)
@@ -387,19 +417,6 @@ class ParsedPartitionSpec:
     return (self.partitions == other.partitions and
             self.sync == other.sync)
 
-  def eq_given_rank(self, other, rank):
-    # ParsedPartitionSpecs may contain trailing empty tuples that don't change
-    # the semantic meaning of the spec but are still valid specs. For example,
-    # for a 2D array, (), ((),), and ((), ()) are all valid specs meaning the
-    # array is fully replicated (no dimension partitioned). This method compares
-    # two specs for semantic equivalence and asserts they are valid specs for
-    # the given array rank.
-    assert len(self.partitions) <= rank and len(other.partitions) <= rank
-    min_length = min(len(self.partitions), len(other.partitions))
-    return (self.partitions[:min_length] == other.partitions[:min_length] and
-            all(p == () for p in self.partitions[min_length:]) and
-            all(p == () for p in other.partitions[min_length:]))
-
   def __len__(self):
     return len(self.partitions)
 
@@ -417,28 +434,61 @@ class ParsedPartitionSpec:
 REPLICATED = ParsedPartitionSpec(None, ())
 
 
-def _prepare_axis_resources(axis_resources, arg_name):
+class CanonicalizedParsedPartitionSpec(ParsedPartitionSpec):
+  """ParsedPartitionSpecs that are canonicalized.
+
+  ParsedPartitionSpecs may contain trailing empty tuples, that make them
+  semantically different in general, and yet in some situations we prefer
+  to regard them as equivalent. For example, partitions of () and ((),)
+  cannot be always considered equivalent, since the first one is a valid
+  spec for a scalar value, while the second is not! However, when either of
+  those are applied to a 2D array, they both mean that the array is fully
+  replicated.
+
+  So CanonicalizedParsedPartitionSpecs removes the trailing empty tuples from
+  partitions.
+  """
+
+  def __init__(self, parsed_pspec: ParsedPartitionSpec):
+    partitions = list(parsed_pspec.partitions)
+    while partitions and partitions[-1] == ():
+      partitions.pop()
+
+    super().__init__(parsed_pspec.unsafe_user_spec, partitions,
+                     parsed_pspec.sync)
+
+  def __repr__(self):
+    return (f"CanonicalizedParsedPartitionSpec(partitions={self.partitions}, "
+            f"unsafe_user_spec={self.unsafe_user_spec}, "
+            f"sync={self.sync})")
+
+
+def _prepare_axis_resources(axis_resources,
+                            arg_name,
+                            allow_unconstrained_dims=False):
   # PyTrees don't treat None values as leaves, so we explicitly need
   # to explicitly declare them as such
   entries, treedef = tree_flatten(axis_resources, is_leaf=lambda x: x is None)
   what = f"{arg_name} leaf specifications"
   entries = [
-      entry if entry is FROM_GDA else ParsedPartitionSpec.from_user_input(
-          entry, what) for entry in entries
+      entry if _is_from_gda(entry) else ParsedPartitionSpec.from_user_input(
+          entry, what, allow_unconstrained_dims=allow_unconstrained_dims)
+      for entry in entries
   ]
   _check_unique_resources(entries, arg_name)
   return tree_unflatten(treedef, entries), entries, treedef
 
 def _check_resources_mismatch(in_axis_resources_flat, is_gda):
-  if not is_gda and in_axis_resources_flat is FROM_GDA:
+  if not is_gda and _is_from_gda(in_axis_resources_flat):
     raise ValueError('For a non-GDA input, the corresponding resource in '
                      'in_axis_resources cannot be `pjit.FROM_GDA`.')
 
 def _check_unique_resources(axis_resources, arg_name):
   for arg_axis_resources in axis_resources:
     if not arg_axis_resources: continue
-    if arg_axis_resources is FROM_GDA: continue
-    resource_counts = Counter(it.chain.from_iterable(arg_axis_resources))
+    if _is_from_gda(arg_axis_resources): continue
+    constrained_dims = [d for d in arg_axis_resources if d is not None]
+    resource_counts = Counter(it.chain.from_iterable(constrained_dims))
     if not resource_counts: continue
     if resource_counts.most_common(1)[0][1] > 1:
       multiple_uses = [r for r, c in resource_counts.items() if c > 1]
@@ -447,10 +497,11 @@ def _check_unique_resources(axis_resources, arg_name):
                          f"to at most one positional dimension, but {arg_axis_resources.user_spec} "
                          f"has duplicate entries for {maps.show_axes(multiple_uses)}")
 
-def _check_shapes_against_resources(what: str, is_global_shape: bool, mesh_shape, flat_avals, flat_axis_resources):
+def _check_shapes_against_resources(what: str, is_global_shape: bool, mesh_shape,
+                                    flat_avals, flat_axis_resources):
   global_str = " global" if is_global_shape else ""
   for aval, aval_axis_resources in zip(flat_avals, flat_axis_resources):
-    if aval_axis_resources is FROM_GDA:
+    if _is_from_gda(aval_axis_resources):
       continue
     shape = aval.shape
     if len(shape) < len(aval_axis_resources):
@@ -459,6 +510,8 @@ def _check_shapes_against_resources(what: str, is_global_shape: bool, mesh_shape
                        f"it has a rank of at least {len(aval_axis_resources)}, "
                        f"but it is {len(shape)}")
     for i, axis_resources in enumerate(aval_axis_resources):
+      if axis_resources is None:
+        continue
       try:
         size = int(np.prod([mesh_shape[resource] for resource in axis_resources], dtype=np.int64))
       except KeyError as e:
@@ -493,12 +546,16 @@ pjit_p.def_impl(_pjit_call_impl)
 @cache()
 def _pjit_lower(
     jaxpr: core.ClosedJaxpr,
-    in_axis_resources: Tuple[ParsedPartitionSpec, ...],
-    out_axis_resources: Tuple[ParsedPartitionSpec, ...],
+    in_axis_resources: Tuple[CanonicalizedParsedPartitionSpec, ...],
+    out_axis_resources: Tuple[CanonicalizedParsedPartitionSpec, ...],
     resource_env,
     donated_invars,
     name: str,
     in_positional_semantics, out_positional_semantics):
+  # in_axis_resources and out_axis_resources are canonicalized to avoid
+  # recompilation (since pjit_lower is cached) if its compiled with `None` but
+  # in the next call `P(None)` is passed. Those are the same thing so should be
+  # treat as equivalent and pjit_lower's cache shouldn't be invalidated.
   in_axes = [get_array_mapping(axes) for axes in in_axis_resources]
   out_axes = [get_array_mapping(axes) for axes in out_axis_resources]
   pxla.resource_typecheck(jaxpr, resource_env, {}, lambda: "pjit")
@@ -772,7 +829,8 @@ ad.reducing_transposes[pjit_p] = _pjit_transpose
 
 
 def _check_resources_against_named_axes(what, aval, pos_axis_resources, named_axis_resources):
-  pjit_resources = set(it.chain.from_iterable(pos_axis_resources))
+  pjit_resources = set(
+      it.chain.from_iterable([d for d in pos_axis_resources if d is not None]))
   aval_resources = set(it.chain.from_iterable(
     named_axis_resources[a] for a in aval.named_shape))
   overlap = pjit_resources & aval_resources
@@ -805,7 +863,8 @@ pxla.custom_resource_typing_rules[pjit_p] = _resource_typing_pjit
 
 def with_sharding_constraint(x, axis_resources):
   x_flat, tree = tree_flatten(x)
-  parsed_axis_resources, entries, _ = _prepare_axis_resources(axis_resources, "axis_resources")
+  parsed_axis_resources, entries, _ = _prepare_axis_resources(
+      axis_resources, "axis_resources", allow_unconstrained_dims=True)
   axis_resources_flat = tuple(
       flatten_axes("with_sharding_constraint axis_resources",
                    tree, parsed_axis_resources))
@@ -837,16 +896,25 @@ def _sharding_constraint_translation_rule(ctx, avals_in, avals_out, x_node, *,
                                           axis_resources, resource_env):
   aval, = avals_in
   mesh = resource_env.physical_mesh
-  return [xla.set_sharding_proto(
-      ctx.builder, x_node, get_aval_sharding_proto(aval, axis_resources, mesh))]
+  return [
+      xla.set_sharding_proto(
+          ctx.builder,
+          x_node,
+          get_aval_sharding_proto(aval, axis_resources, mesh),
+          unspecified_dims=get_unconstrained_dims(axis_resources))
+  ]
 xla.register_translation(sharding_constraint_p, _sharding_constraint_translation_rule)
 
-def _sharding_constraint_mhlo_lowering(ctx, avals_in, avals_out, x_node, *,
-                                       axis_resources, resource_env):
-  aval, = avals_in
+def _sharding_constraint_mhlo_lowering(ctx, x_node, *, axis_resources,
+                                       resource_env):
+  aval, = ctx.avals_in
   mesh = resource_env.physical_mesh
-  return [mlir.wrap_with_sharding_op(
-      x_node, get_aval_sharding_proto(aval, axis_resources, mesh))]
+  return [
+      mlir.wrap_with_sharding_op(
+          x_node,
+          get_aval_sharding_proto(aval, axis_resources, mesh),
+          unspecified_dims=get_unconstrained_dims(axis_resources))
+  ]
 mlir.register_lowering(sharding_constraint_p,
                        _sharding_constraint_mhlo_lowering)
 
@@ -877,7 +945,7 @@ pxla.custom_resource_typing_rules[sharding_constraint_p] = \
 def get_array_mapping(axis_resources: ParsedPartitionSpec) -> pxla.ArrayMapping:
   return OrderedDict((axis, i)
                      for i, axes in enumerate(axis_resources)
-                     for axis in axes)
+                     if axes is not None for axis in axes)
 
 
 def get_aval_sharding_proto(aval: core.AbstractValue,
@@ -888,42 +956,53 @@ def get_aval_sharding_proto(aval: core.AbstractValue,
       aval, array_mapping)
   return sharding_spec.sharding_proto()
 
+
+def get_unconstrained_dims(axis_resources: ParsedPartitionSpec):
+  return {i for i, axes in enumerate(axis_resources) if axes is None}
+
+
 def global_to_local(positional_semantics, mesh, avals, axes):
   if isinstance(positional_semantics, maps._PositionalSemantics):
     positional_semantics = [positional_semantics] * len(axes)
   return [
-      aval if ps == maps._PositionalSemantics.GLOBAL else mesh.global_to_local(
+      aval if ps == maps._PositionalSemantics.GLOBAL or aval_axes.partitions == () else mesh.global_to_local(
           get_array_mapping(aval_axes), aval)
       for aval, aval_axes, ps in safe_zip(avals, axes, positional_semantics)
   ]
 
 def local_to_global(positional_semantics, mesh, avals, axes):
   return [
-      aval if ps == maps._PositionalSemantics.GLOBAL else mesh.local_to_global(
+      aval if ps == maps._PositionalSemantics.GLOBAL or aval_axes.partitions == () else mesh.local_to_global(
           get_array_mapping(aval_axes), aval)
       for aval, aval_axes, ps in safe_zip(avals, axes, positional_semantics)
   ]
 
-def _canonicalize_spec(in_axis_resources_flat: ParsedPartitionSpec, arg):
+def _create_cpspec(x):
+  return x if _is_from_gda(x) else CanonicalizedParsedPartitionSpec(x)
+
+def _maybe_replace_from_gda_with_pspec(
+    in_axis_resources_flat: CanonicalizedParsedPartitionSpec, arg) -> CanonicalizedParsedPartitionSpec:
   if isinstance(arg, GDA):
-    gsda_ppspec = gsda_mesh_axes_to_parsed_pspec(arg._mesh_axes)
-    if (in_axis_resources_flat is not FROM_GDA and
-        not in_axis_resources_flat.eq_given_rank(gsda_ppspec, len(arg.shape))):
+    gda_cpspec = gda_mesh_axes_to_canonicalized_parsed_pspec(arg._mesh_axes)
+    assert type(gda_cpspec) is CanonicalizedParsedPartitionSpec
+    if (not _is_from_gda(in_axis_resources_flat) and
+        in_axis_resources_flat != gda_cpspec):
       raise ValueError(
           'Got an input GDA to pjit with different partitioning than specified in '
           "the in_axis_resources argument to pjit. The partitioning must match, or "
           "use `jax.experimental.pjit.FROM_GDA` in `in_axis_resources`. "
-          f"Got GDA spec: {gsda_ppspec.user_spec} and "
+          f"Got GDA spec: {gda_cpspec.user_spec} and "
           f"pjit spec: {in_axis_resources_flat.user_spec} for GDA: {arg}")
-    return gsda_ppspec
+    return gda_cpspec
   return in_axis_resources_flat
 
-def gsda_mesh_axes_to_parsed_pspec(mesh_axes) -> ParsedPartitionSpec:
+def gda_mesh_axes_to_canonicalized_parsed_pspec(mesh_axes) -> CanonicalizedParsedPartitionSpec:
   if not isinstance(mesh_axes, PartitionSpec):
     pspec = PartitionSpec(*mesh_axes)
   else:
     pspec = mesh_axes
-  return ParsedPartitionSpec.from_user_input(pspec, arg_name='GDA mesh_axes')
+  return CanonicalizedParsedPartitionSpec(ParsedPartitionSpec.from_user_input(
+      pspec, arg_name='GDA mesh_axes'))
 
 # -------------------- XLA OpSharding to PartitionSpec --------------------
 # Note that OpSharding is more expressive than PartitionSpecs, so it's not
