@@ -218,7 +218,8 @@ def backward_pass(jaxpr: core.Jaxpr, reduce_axes, consts, primals_in, cotangents
     with source_info_util.user_context(eqn.source_info.traceback):
       if eqn.primitive.call_primitive or eqn.primitive.map_primitive:
         cts_in_avals = [v.aval for v in eqn.outvars]
-        call_jaxpr, params = core.extract_call_jaxpr(eqn.primitive, eqn.params)
+        params = dict(eqn.params)
+        call_jaxpr = params.pop('call_jaxpr')
         cts_out = get_primitive_transpose(eqn.primitive)(
             params, call_jaxpr, invals, cts_in, cts_in_avals, reduce_axes)
       elif eqn.primitive in reducing_transposes:
@@ -358,7 +359,7 @@ class JVPTrace(Trace):
     tangents_out = map(recast_to_float0, primals_out, tangents_out)
     return map(partial(JVPTracer, self), primals_out, tangents_out)
 
-  def post_process_custom_jvp_call(self, out_tracers, params):
+  def post_process_custom_jvp_call(self, out_tracers, _):
     raise CustomJVPException()
 
   def process_custom_vjp_call(self, _, __, fwd, bwd, tracers, *, out_trees):
@@ -374,7 +375,7 @@ class JVPTrace(Trace):
     tangents_out = map(recast_to_float0, primals_out, tangents_out)
     return map(partial(JVPTracer, self), primals_out, tangents_out)
 
-  def post_process_custom_vjp_call(self, out_tracers, params):
+  def post_process_custom_vjp_call(self, out_tracers, _):
     raise CustomVJPException()
 
   def join(self, xt, yt):
