@@ -17,11 +17,14 @@
 
 import collections
 import dataclasses
+import functools
 from functools import partial
 import io
+import itertools
+import re
 import typing
-from typing import (Any, Callable, Dict, List, Optional, Sequence, Type, Union,
-                    Tuple)
+from typing import (Any, Callable, Dict, List, Optional, Sequence, Set, Tuple,
+                    Type, Union)
 from typing_extensions import Protocol
 import warnings
 
@@ -36,6 +39,7 @@ from jax._src.lib.mlir.dialects import chlo
 from jax._src.lib.mlir.dialects import mhlo
 from jax._src.lib.mlir.dialects import std
 from jax._src.lib import xla_client as xc
+import jax._src.pretty_printer as pp
 from jax._src import source_info_util
 import jax._src.util as util
 import jax.interpreters.ad as ad
@@ -159,7 +163,7 @@ def ir_constants(val: Any,
   Returns:
     A representation of the constant as a list of IR values.
   """
-  for t in type(val).mro():
+  for t in type(val).__mro__:
     handler = _constant_handlers.get(t)
     if handler: return handler(val, canonicalize_types)
   if hasattr(val, '__jax_array__'):
@@ -258,18 +262,28 @@ for t in device_array.device_array_types:
 # Source locations
 
 def _source_info_to_location(
-    source_info: source_info_util.SourceInfo) -> ir.Location:
+    primitive: core.Primitive, params: Dict,
+    source_info: source_info_util.SourceInfo,
+    name_stack: str = "") -> ir.Location:
+  eqn_str = str(
+      pp.text(name_stack) +
+      core.pp_eqn_compact(primitive.name, params, core.JaxprPpContext()))
   frame = source_info_util.user_frame(source_info)
   if frame is None:
-    return ir.Location.unknown()
-  return ir.Location.file(xla._get_canonical_source_file(frame), frame.line_num,
-                          1)
+    loc = ir.Location.unknown()
+  else:
+    loc = ir.Location.file(xla._get_canonical_source_file(frame),
+                           frame.line_num, 1)
+  loc = ir.Location.name(eqn_str, childLoc=loc)
+  # TODO(phawkins): also include primitive.name as the operator type.
+  return loc
 
 
 # Translation rules
 
 @dataclasses.dataclass
-class LoweringContext:
+class ModuleContext:
+  """Module-wide context information for MLIR lowering."""
   context: ir.Context
   module: ir.Module
   ip: ir.InsertionPoint
@@ -278,12 +292,16 @@ class LoweringContext:
   axis_env: xla.AxisEnv
   name_stack: str
 
+  # Cached primitive lowerings.
+  cached_primitive_lowerings: Dict[Any, builtin.FuncOp]
 
-  def __init__(self, platform: str, axis_env: xla.AxisEnv, name_stack: str,
-               context: Optional[ir.Context] = None,
-               module: Optional[ir.Module] = None,
-               ip: Optional[ir.InsertionPoint] = None,
-               symbol_table: Optional[ir.SymbolTable] = None):
+  def __init__(
+      self, platform: str, axis_env: xla.AxisEnv, name_stack: str,
+      context: Optional[ir.Context] = None,
+      module: Optional[ir.Module] = None,
+      ip: Optional[ir.InsertionPoint] = None,
+      symbol_table: Optional[ir.SymbolTable] = None,
+      cached_primitive_lowerings: Optional[Dict[Any, builtin.FuncOp]] = None):
     assert platform is not None
     self.context = context or ir.Context()
     self.module = module or ir.Module.create(loc=ir.Location.unknown(self.context))
@@ -292,17 +310,28 @@ class LoweringContext:
     self.platform = platform
     self.axis_env = axis_env
     self.name_stack = name_stack
+    self.cached_primitive_lowerings = ({} if cached_primitive_lowerings is None
+                                       else cached_primitive_lowerings)
     mhlo.register_mhlo_dialect(self.context)
     chlo.register_chlo_dialect(self.context)
 
   def replace(self, **kw): return dataclasses.replace(self, **kw)
 
 
+@dataclasses.dataclass
+class LoweringRuleContext:
+  """Per-rule context information for MLIR lowering."""
+  module_context: ModuleContext
+  primitive: Optional[core.Primitive]
+  avals_in: Sequence[core.AbstractValue]
+  avals_out: Any  # Usually Sequence[core.AbstractValue], but sometimes None.
+
+  def replace(self, **kw): return dataclasses.replace(self, **kw)
+
+
 if not MYPY:
   class LoweringRule(Protocol):
-    def __call__(self, ctx: LoweringContext,
-                 avals_in: Sequence[core.AbstractValue],
-                 avals_out: Sequence[core.AbstractValue],
+    def __call__(self, ctx: LoweringRuleContext,
                  *args: Union[ir.Value, Sequence[ir.Value]],
                  **kw) -> Sequence[Union[ir.Value, Sequence[ir.Value]]]:
       """Converts a JAX primitive invocation into MLIR."""
@@ -331,20 +360,24 @@ def flatten_lowering_ir_args(
 ) -> Sequence[Sequence[ir.Value]]:
   return util.flatten(map(wrap_singleton_ir_values, xs))
 
+_module_unique_id = itertools.count()
+_module_name_regex = re.compile(r"[^\w.-]")
+
 def lower_jaxpr_to_module(
-    jaxpr: core.ClosedJaxpr, platform: str, axis_env: xla.AxisEnv,
+    module_name: str, jaxpr: core.ClosedJaxpr, platform: str,
+    axis_env: xla.AxisEnv,
     name_stack: str, donated_args: Sequence[bool],
     replicated_args: Optional[Sequence[bool]] = None,
     arg_shardings: Optional[Sequence[Optional[xc.OpSharding]]] = None,
     result_shardings: Optional[Sequence[Optional[xc.OpSharding]]] = None
-    ) -> str:
+    ) -> ir.Module:
   """Lowers a top-level jaxpr to an MHLO module.
 
   Handles the quirks of the argument/return value passing conventions of the
   runtime."""
   input_output_aliases = None
   if platform in ("gpu", "tpu"):
-    input_output_aliases, out_donated_args = _set_up_aliases(
+    input_output_aliases, donated_args = _set_up_aliases(
         jaxpr.in_avals, jaxpr.out_avals, donated_args)
   if any(donated_args):
     # TODO(tomhennigan): At call time we should mark these buffers as deleted.
@@ -353,8 +386,15 @@ def lower_jaxpr_to_module(
     warnings.warn("Some donated buffers were not usable: {}".format(
         ", ".join(unused_donations)))
 
-  ctx = LoweringContext(platform, axis_env, name_stack)
+  ctx = ModuleContext(platform, axis_env, name_stack)
   with ctx.context, ir.Location.unknown(ctx.context):
+    # Remove module name characters that XLA would alter. This ensures that
+    # XLA computation preserves the module name.
+    module_name = _module_name_regex.sub("_", module_name)
+    # Some clients expect modules to have unique names, e.g., in trace data.
+    # This may or may not be a reasonable assumption.
+    ctx.module.operation.attributes["sym_name"] = ir.StringAttr.get(
+        f"{module_name}.{next(_module_unique_id)}")
     # TODO(phawkins): represent units with zero buffers at the runtime level.
     lower_jaxpr_to_fun(
         ctx, "main", jaxpr, public=True, replace_units_with_dummy=True,
@@ -362,11 +402,13 @@ def lower_jaxpr_to_module(
         arg_shardings=arg_shardings, result_shardings=result_shardings,
         input_output_aliases=input_output_aliases)
 
-
   ctx.module.operation.verify()
+  return ctx.module
+
+def module_to_string(module: ir.Module) -> str:
   output = io.StringIO()
-  ctx.module.operation.print(file=output, enable_debug_info=True,
-                             print_generic_op_form=False)
+  module.operation.print(file=output, enable_debug_info=True,
+                         print_generic_op_form=False)
   return output.getvalue()
 
 def _set_up_aliases(avals_in, avals_out, donated_args):
@@ -387,7 +429,7 @@ def _set_up_aliases(avals_in, avals_out, donated_args):
   return input_output_aliases, out_donated_args
 
 def lower_jaxpr_to_fun(
-    ctx: LoweringContext, name: str, jaxpr: core.ClosedJaxpr, *,
+    ctx: ModuleContext, name: str, jaxpr: core.ClosedJaxpr, *,
     public: bool = False, replace_units_with_dummy: bool = False,
     replace_tokens_with_dummy: bool = False,
     replicated_args: Optional[Sequence[bool]] = None,
@@ -509,8 +551,27 @@ def lower_jaxpr_to_fun(
 
   return func_op
 
+def _emit_lowering_rule_as_fun(lowering_rule,
+                               ctx: LoweringRuleContext) -> builtin.FuncOp:
+  """Emits the contents of a lowering rule as a private function."""
+  input_types = map(aval_to_ir_types, ctx.avals_in)
+  output_types = map(aval_to_ir_types, ctx.avals_out)
+  flat_input_types = util.flatten(input_types)
+  flat_output_types = util.flatten(output_types)
+  ftype = ir.FunctionType.get(flat_input_types, flat_output_types)
+  assert ctx.primitive is not None
+  func_op = builtin.FuncOp(ctx.primitive.name, ftype, ip=ctx.module_context.ip)
+  func_op.attributes["sym_visibility"] = ir.StringAttr.get("private")
+  ctx.module_context.symbol_table.insert(func_op)
+  entry_block = func_op.add_entry_block()
+  with ir.InsertionPoint(entry_block):
+    unflattened_args = util.unflatten(entry_block.arguments,
+                                      map(len, input_types))
+    outs = lowering_rule(ctx, *_unwrap_singleton_ir_values(unflattened_args))
+    std.ReturnOp(util.flatten(map(wrap_singleton_ir_values, outs)))
+  return func_op
 
-def jaxpr_subcomp(ctx: LoweringContext, jaxpr: core.Jaxpr,
+def jaxpr_subcomp(ctx: ModuleContext, jaxpr: core.Jaxpr,
                   consts: Sequence[Sequence[ir.Value]],
                   *args: Sequence[ir.Value]) -> Sequence[Sequence[ir.Value]]:
   """Lowers a jaxpr into mHLO, inlined into an existing function.
@@ -543,9 +604,8 @@ def jaxpr_subcomp(ctx: LoweringContext, jaxpr: core.Jaxpr,
   map(write, jaxpr.invars, args)
   for eqn in jaxpr.eqns:
     in_nodes = map(read, eqn.invars)
-    # TODO(phawkins): attach the primitive name, parameters, and name stack as
-    # metadata.
-    loc = _source_info_to_location(eqn.source_info)
+    loc = _source_info_to_location(eqn.primitive, eqn.params, eqn.source_info,
+                                   name_stack=ctx.name_stack)
     with source_info_util.user_context(eqn.source_info.traceback), loc:
       if eqn.primitive in _platform_specific_lowerings[ctx.platform]:
         rule = _platform_specific_lowerings[ctx.platform][eqn.primitive]
@@ -553,14 +613,16 @@ def jaxpr_subcomp(ctx: LoweringContext, jaxpr: core.Jaxpr,
         rule = _lowerings[eqn.primitive]
       elif (eqn.primitive in xla._translations or
             eqn.primitive in xla._backend_specific_translations[ctx.platform]):
-        rule = partial(xla_fallback_lowering, eqn.primitive)
+        rule = xla_fallback_lowering(eqn.primitive)
       else:
         raise NotImplementedError(
             f"MLIR translation rule for primitive '{eqn.primitive.name}' not "
             f"found for platform {ctx.platform}")
 
-      ans = rule(ctx, map(aval, eqn.invars), map(aval, eqn.outvars),
-                 *map(_unwrap_singleton_ir_values, in_nodes),
+      rule_ctx = LoweringRuleContext(
+          module_context=ctx, primitive=eqn.primitive,
+          avals_in=map(aval, eqn.invars), avals_out=map(aval, eqn.outvars))
+      ans = rule(rule_ctx, *map(_unwrap_singleton_ir_values, in_nodes),
                  **eqn.params)
 
     try:
@@ -586,15 +648,16 @@ def lower_fun(fun: Callable, multiple_results: bool = True) -> Callable:
 
   The returned function does not use `avals_out`, so callers may pass any value
   as `avals_out`."""
-  def f_lowered(ctx, avals_in, avals_out, *args, **params):
+  def f_lowered(ctx, *args, **params):
     if multiple_results:
       f = fun
     else:
       f = lambda *args, **kw: (fun(*args, **kw),)
     wrapped_fun = lu.wrap_init(f, params)
-    with core.extend_axis_env_nd(zip(ctx.axis_env.names, ctx.axis_env.sizes)):
-      jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(wrapped_fun, avals_in)
-    return jaxpr_subcomp(ctx, jaxpr, _ir_consts(consts),
+    axis_env = ctx.module_context.axis_env
+    with core.extend_axis_env_nd(zip(axis_env.names, axis_env.sizes)):
+      jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(wrapped_fun, ctx.avals_in)
+    return jaxpr_subcomp(ctx.module_context, jaxpr, _ir_consts(consts),
                          *map(wrap_singleton_ir_values, args))
 
   return f_lowered
@@ -615,19 +678,20 @@ def _call_lowering(fn_name, stack_name, call_jaxpr, backend, ctx, avals_in,
                     flatten_lowering_ir_args(args))
   return util.unflatten(call.results, map(len, output_types))
 
-def _xla_call_lower(ctx, avals_in, avals_out, *args,
+def _xla_call_lower(ctx, *args,
                     backend=None, name, call_jaxpr, donated_invars, inline=None,
                     device=None):
   del device, donated_invars, inline  # Ignored.
   return _call_lowering(f"jit_{name}", xla.wrap_name(name, "jit"), call_jaxpr,
-                        backend, ctx, avals_in, avals_out, *args)
+                        backend, ctx.module_context, ctx.avals_in, ctx.avals_out,
+                        *args)
 
 register_lowering(xla.xla_call_p, _xla_call_lower)
 
-def _named_call_lowering(ctx, avals_in, avals_out, *args, name, backend=None,
+def _named_call_lowering(ctx, *args, name, backend=None,
                          call_jaxpr):
-  return _call_lowering(name, name, call_jaxpr, backend, ctx, avals_in,
-                        avals_out, *args)
+  return _call_lowering(name, name, call_jaxpr, backend, ctx.module_context,
+                        ctx.avals_in, ctx.avals_out, *args)
 
 register_lowering(core.named_call_p, _named_call_lowering)
 register_lowering(core.call_p, partial(_named_call_lowering, name="core_call"))
@@ -639,18 +703,17 @@ def full_like_aval(value, aval: core.ShapedArray) -> ir.Value:
   return mhlo.BroadcastOp(aval_to_ir_type(aval), zero,
                           dense_int_elements(aval.shape)).result
 
-def zeros_like_lowering(ctx, avals_in, avals_out, x):
-  aval, = avals_in
+def zeros_like_lowering(ctx, x):
+  aval, = ctx.avals_in
   assert isinstance(aval, core.ShapedArray), aval
   return [full_like_aval(0, aval)]
 register_lowering(ad_util.zeros_like_p, zeros_like_lowering)
 
-def add_jaxvals_lowering(ctx, avals_in, avals_out, x, y):
+def add_jaxvals_lowering(ctx, x, y):
   return mhlo.AddOp(x, y).results
 register_lowering(ad_util.add_jaxvals_p, add_jaxvals_lowering)
 
-register_lowering(ad_util.stop_gradient_p,
-                  lambda ctx, avals_in, avals_out, x: [x])
+register_lowering(ad_util.stop_gradient_p, lambda ctx, x: [x])
 
 
 def _minmax_mhlo(op, cmp, x, y):
@@ -697,11 +760,20 @@ def convert_mhlo(x, aval_in, aval_out):
   return mhlo.ConvertOp(aval_to_ir_type(aval_out), x).result
 
 
-def wrap_with_sharding_op(x, sharding_proto: xc.OpSharding):
+def wrap_with_sharding_op(x,
+                          sharding_proto: xc.OpSharding,
+                          unspecified_dims: Optional[Set[int]] = None):
+  # unspecified_dims indicate dimensions whose shardings are not specified and
+  # XLA sharding propagation can change them.
+  if unspecified_dims:
+    backend_config = "unspecified_dims=[" + ",".join(
+        [str(i) for i in sorted(unspecified_dims)]) + "]"
+  else:
+    backend_config = ""
   op = mhlo.CustomCallOp([x.type], [x],
                          call_target_name=ir.StringAttr.get("Sharding"),
                          has_side_effect=ir.BoolAttr.get(False),
-                         backend_config=ir.StringAttr.get(""),
+                         backend_config=ir.StringAttr.get(backend_config),
                          api_version=i32_attr(1),
                          called_computations=ir.ArrayAttr.get([]),
                          operand_layouts=None,
@@ -716,122 +788,73 @@ def set_sharding(op, sharding_proto: xc.OpSharding):
 
 # MLIR lowerings for lax primitives
 
-def xla_fallback_lowering(prim: core.Primitive, ctx: LoweringContext,
-                          avals_in, avals_out, *args, **params):
-  xla_computation = xla.primitive_subcomputation(
-      ctx.platform, ctx.axis_env, prim, *avals_in, **params)
-  submodule_str = xc._xla.mlir.xla_computation_to_mlir_module(xla_computation)
-  submodule = ir.Module.parse(submodule_str)
-  callee_name = None
-  for op in submodule.body.operations:
-    ctx.module.body.append(op)
-    if op.name.value == "main":
-      op.attributes["sym_name"] = ir.StringAttr.get(f"xla_fallback_{prim.name}")
-      callee_name = ir.StringAttr(ctx.symbol_table.insert(op)).value
-      op.attributes["sym_visibility"] = ir.StringAttr.get("private")
-    else:
-      ctx.symbol_table.insert(op)
+def cache_lowering(f):
+  """Decorator that causes the contents of a lowering rule to be reused.
 
-  output_types = map(aval_to_ir_types, avals_out)
-  flat_output_types = util.flatten(output_types)
-  output_type = (ir.TupleType.get_tuple(flat_output_types)
-                 if prim.multiple_results else flat_output_types[0])
+  The lowering will be emitted out-of-line in a separate function, together with
+  a call to that function. If the same primitive is called with the same shapes
+  and parameters, a new call to the original function will be added, without
+  emitting a new function.
+  """
+  @functools.wraps(f)
+  def cached_lowering(ctx, *args, **params):
+    assert ctx.primitive is not None
+    key = (ctx.primitive, tuple(ctx.avals_in), tuple(ctx.avals_out),
+           tuple(params.items()))
+    try:
+      func = ctx.module_context.cached_primitive_lowerings.get(key)
+    except TypeError:
+      # If the parameters aren't hashable, give up on caching.
+      # TODO(phawkins): switch to requiring hashability, when XLA fallback
+      # computations have been ported to MHLO.
+      return f(ctx, *args, **params)
+    if func is None:
+      func = _emit_lowering_rule_as_fun(partial(f, **params), ctx)
+      ctx.module_context.cached_primitive_lowerings[key] = func
 
-  call = std.CallOp([output_type], ir.FlatSymbolRefAttr.get(callee_name),
-                    flatten_lowering_ir_args(args)).result
-  if not prim.multiple_results:
-    return [call]
-  flat_results = [mhlo.GetTupleElementOp(typ, call, i32_attr(i)).result
-                  for i, typ in enumerate(flat_output_types)]
-  return util.unflatten(flat_results, map(len, output_types))
+    output_types = map(aval_to_ir_types, ctx.avals_out)
+    flat_output_types = util.flatten(output_types)
+    call = std.CallOp(flat_output_types,
+                      ir.FlatSymbolRefAttr.get(func.name.value),
+                      flatten_lowering_ir_args(args))
+    return util.unflatten(call.results, map(len, output_types))
+  return cached_lowering
 
 
+def xla_fallback_lowering(prim: core.Primitive):
+  @cache_lowering
+  def fallback(ctx: LoweringRuleContext, *args, **params):
+    module_ctx = ctx.module_context
+    xla_computation = xla.primitive_subcomputation(
+        module_ctx.platform, module_ctx.axis_env, prim, *ctx.avals_in, **params)
+    submodule_str = xc._xla.mlir.xla_computation_to_mlir_module(xla_computation)
+    submodule = ir.Module.parse(submodule_str)
+    callee_name = None
+    for op in submodule.body.operations:
+      module_ctx.module.body.append(op)
+      if op.name.value == "main":
+        op.attributes["sym_name"] = ir.StringAttr.get(f"xla_fallback_{prim.name}")
+        callee_name = ir.StringAttr(module_ctx.symbol_table.insert(op)).value
+        op.attributes["sym_visibility"] = ir.StringAttr.get("private")
+      else:
+        module_ctx.symbol_table.insert(op)
+
+    output_types = map(aval_to_ir_types, ctx.avals_out)
+    flat_output_types = util.flatten(output_types)
+    output_type = (ir.TupleType.get_tuple(flat_output_types)
+                   if prim.multiple_results else flat_output_types[0])
+
+    call = std.CallOp([output_type], ir.FlatSymbolRefAttr.get(callee_name),
+                      flatten_lowering_ir_args(args)).result
+    if not prim.multiple_results:
+      return [call]
+    flat_results = [mhlo.GetTupleElementOp(typ, call, i32_attr(i)).result
+                    for i, typ in enumerate(flat_output_types)]
+    return util.unflatten(flat_results, map(len, output_types))
+  return fallback
+
 register_lowering(ad.custom_lin_p, ad._raise_custom_vjp_error_on_jvp)
 
-
-def _dummy_like_aval(aval: core.AbstractValue) -> Sequence[ir.Value]:
-  if isinstance(aval, core.ShapedArray):
-    return [full_like_aval(0, aval)]
-  elif isinstance(aval, core.AbstractToken):
-    return mhlo.CreateTokenOp(aval_to_ir_type(aval)).results
-  elif isinstance(aval, core.AbstractUnit):
-    return ()
-  else:
-    raise TypeError(f"Unsupported abstract value {aval}")
-
-def _remat_using_while(ctx, avals_in, avals_out, *args, name, call_jaxpr):
-  input_types = map(aval_to_ir_types, avals_in)
-  output_types = map(aval_to_ir_types, avals_out)
-  flat_output_types = util.flatten(output_types)
-  int32_scalar_type = aval_to_ir_type(core.ShapedArray((), np.dtype(np.int32)))
-  loop_carry_types = [(int32_scalar_type,)] + input_types + output_types
-  flat_loop_carry_types = util.flatten(loop_carry_types)
-  counter_init = ir_constants(np.array(0, np.int32))
-  flat_args = flatten_lowering_ir_args(
-      (counter_init,) + args +
-      tuple(_dummy_like_aval(aval) for aval in avals_out))
-  loop_carry_tuple_type = ir.TupleType.get_tuple(flat_loop_carry_types)
-  init_carry = mhlo.TupleOp(loop_carry_tuple_type, flat_args)
-
-  one = ir_constant(np.array(1, np.int32))
-  while_op = mhlo.WhileOp([loop_carry_tuple_type], [init_carry.result])
-
-  # Loop condition
-  cond_block = while_op.regions[0].blocks.append(loop_carry_tuple_type)
-  with ir.InsertionPoint(cond_block):
-    bool_scalar_type = aval_to_ir_type(core.ShapedArray((), np.dtype(np.bool_)))
-    two = ir_constant(np.array(2, np.int32))
-    shape = ir_constant(np.array((), np.int64), canonicalize_types=False)
-    rng = mhlo.RngUniformOp(one, two, shape).result
-    i = mhlo.GetTupleElementOp(int32_scalar_type, cond_block.arguments[0],
-                               i32_attr(0))
-    cmp = mhlo.CompareOp(bool_scalar_type, i, rng, ir.StringAttr.get("LT"),
-                         ir.StringAttr.get("SIGNED")).result
-    mhlo.ReturnOp([cmp])
-
-  body_block = while_op.regions[1].blocks.append(loop_carry_tuple_type)
-  with ir.InsertionPoint(body_block):
-    flat_body_args = [
-        mhlo.GetTupleElementOp(input_type, body_block.arguments[0],
-                               i32_attr(i)).result
-        for i, input_type in enumerate(flat_loop_carry_types)]
-    body_args = util.unflatten(flat_body_args, map(len, loop_carry_types))
-    ((i,),), y, _ = util.split_list(body_args, [1, len(avals_in)])
-    body_ctx = ctx.replace(
-        name_stack=xla.extend_name_stack(ctx.name_stack,
-                                         xla.wrap_name(name, 'remat')))
-    z = jaxpr_subcomp(body_ctx, call_jaxpr, (), *y)
-    i_next = mhlo.AddOp(i, one).result
-    new_carry = mhlo.TupleOp(
-        loop_carry_tuple_type,
-        [i_next, *util.flatten(y), *util.flatten(z)])
-    mhlo.ReturnOp([new_carry.result])
-
-  outputs = [mhlo.GetTupleElementOp(output_type, while_op.result,
-                                    i32_attr(1 + len(avals_in) + i)
-                                   ).result
-             for i, output_type in enumerate(flat_output_types)]
-  return util.unflatten(outputs, map(len, output_types))
-
-
-def _remat_lowering(ctx, avals_in, avals_out, *args,
-                    name, call_jaxpr,
-                    prevent_cse, differentiated, concrete,
-                    policy, device=None):
-  del device, concrete, policy  # Unused.
-  if differentiated and prevent_cse:
-    if True: # ctx.platform == "gpu":
-      return _remat_using_while(ctx, avals_in, avals_out, *args, name=name,
-                                call_jaxpr=call_jaxpr)
-    else:
-      assert False
-      #return _remat_using_cond(ctx, args, name, call_jaxpr)
-  else:
-    return jaxpr_subcomp(
-        ctx, call_jaxpr, (), *map(wrap_singleton_ir_values, args))
-
-register_lowering(pe.remat_call_p, _remat_lowering)
-
 # Lax ops missing MLIR lowerings.
 # # TODO(b/203775215): these are missing from the cHLO dialect. Either add
 # # them or port them to Python.
