@@ -57,6 +57,7 @@ from jax._src.util import (unzip3, prod, safe_map, safe_zip,
 from jax.errors import JAXTypeError
 from jax._src import dispatch
 from jax._src import profiler
+from jax._src.lib import _xla_extension_version
 from jax._src.lib import xla_bridge as xb
 from jax._src.lib import xla_client as xc
 from jax._src.lib import pmap_lib
@@ -84,6 +85,7 @@ else:
   OrderedDictType = Dict
 
 xops = xc.ops
+xe = xc._xla
 
 unsafe_map, map = map, safe_map  # type: ignore
 
@@ -314,7 +316,7 @@ def shard_args(devices: Sequence[xb.xla_client.Device],
     A list of length matching args, containing lists of per-device buffers
     for each argument.
   """
-  return [_shard_arg(arg, devices, indices[a]) for a, arg in enumerate(args)]
+  return [_shard_arg(arg, devices, indices[i]) for i, arg in enumerate(args)]
 
 
 shard_arg_handlers: Dict[Any, Callable[[Any, Any, Any], Sequence[Any]]] = {}
@@ -800,6 +802,7 @@ def parallel_callable(fun: lu.WrappedFun,
 
 @dataclasses.dataclass(frozen=True)
 class ParallelCallableInfo:
+  name: str
   backend: xla.Backend
   axis_name: core.AxisName
   axis_size: int
@@ -846,7 +849,7 @@ def find_replicas(jaxpr, axis_size, global_axis_size):
   return ReplicaInfo(jaxpr_replicas, num_local_replicas, num_global_replicas)
 
 
-def tuple_args(shards: ShardInfo):
+def should_tuple_args(shards: ShardInfo):
   # tuplify long arg lists for TPU
   return len(shards.global_sharded_avals) > 100
 
@@ -868,8 +871,10 @@ def stage_parallel_callable(
     global_sharded_avals = sharded_avals  # type: ignore
 
   with core.extend_axis_env(pci.axis_name, pci.global_axis_size, None):  # type: ignore
-    jaxpr, out_sharded_avals, consts = pe.trace_to_jaxpr_final(
-        fun, global_sharded_avals, pe.debug_info_final(fun, "pmap"))
+    with dispatch.log_elapsed_time(f"Finished tracing + transforming {fun.__name__} "
+                                   "for pmap in {elapsed_time} sec"):
+      jaxpr, out_sharded_avals, consts = pe.trace_to_jaxpr_final(
+          fun, global_sharded_avals, pe.debug_info_final(fun, "pmap"))
   jaxpr = dispatch.apply_outfeed_rewriter(jaxpr)
 
   assert len(out_sharded_avals) == len(pci.out_axes), (
@@ -959,8 +964,8 @@ def lower_parallel_callable(
       must_run_on_all_devices = True
 
   pci = ParallelCallableInfo(
-      backend, axis_name, axis_size, global_axis_size, devices, in_axes,
-      out_axes_thunk, avals)
+      name, backend, axis_name, axis_size, global_axis_size, devices,
+      in_axes, out_axes_thunk, avals)
   jaxpr, consts, replicas, parts, shards = stage_parallel_callable(
       pci, fun, global_arg_shapes)
 
@@ -1016,36 +1021,48 @@ def lower_parallel_callable(
   closed_jaxpr = core.ClosedJaxpr(jaxpr, consts)
   replicated_args = [axis is None for axis in in_axes]
   module: Union[str, xc.XlaComputation]
+  tuple_args = should_tuple_args(shards)
+  module_name = f"pmap_{fun.__name__}"
   with maybe_extend_axis_env(axis_name, global_axis_size, None):  # type: ignore
     if config.jax_enable_mlir:
       module = mlir.lower_jaxpr_to_module(
-          closed_jaxpr, backend.platform, axis_env, name_stack, donated_invars,
-          replicated_args=replicated_args,
+          module_name, closed_jaxpr, backend.platform, axis_env,
+          name_stack, donated_invars, replicated_args=replicated_args,
           arg_shardings=_shardings_to_mlir_shardings(parts.arg_parts),
           result_shardings=_shardings_to_mlir_shardings(parts.out_parts))
     else:
       module = xla.lower_jaxpr_to_xla_module(
-          f"pmap_{fun.__name__}", closed_jaxpr, backend.platform, axis_env,
-          name_stack, tuple_args(shards), donated_invars, replicated_args,
+          module_name, closed_jaxpr, backend.platform, axis_env,
+          name_stack, tuple_args, donated_invars, replicated_args,
           parts.arg_parts, parts.out_parts)
-
-  return PmapComputation(module, pci, replicas, parts, shards)
+  return PmapComputation(module, pci=pci, replicas=replicas, parts=parts,
+                         shards=shards, tuple_args=tuple_args)
 
 
 class PmapComputation:
-  def __init__(self, hlo, *compile_args):
+  def __init__(self, hlo, **compile_args):
     self._executable = None
     self._hlo = hlo
     self.compile_args = compile_args
 
   def hlo(self):
     # this is a method for api consistency with dispatch.XlaComputation
-    return self._hlo
+    if isinstance(self._hlo, xc.XlaComputation):
+      return self._hlo
+    else:
+      return xe.mlir.mlir_module_to_xla_computation(
+          mlir.module_to_string(self._hlo),
+          use_tuple_args=self.compile_args["tuple_args"])
+
+  def mhlo(self) -> str:
+    if isinstance(self._hlo, xc.XlaComputation):
+      return xe.mlir.xla_computation_to_mlir_module(self._hlo)
+    return mlir.module_to_string(self._hlo)
 
   @profiler.annotate_function
   def compile(self):
     if self._executable is None:
-      self._executable = PmapExecutable.from_hlo(self._hlo, *self.compile_args)
+      self._executable = PmapExecutable.from_hlo(self._hlo, **self.compile_args)
     return self._executable
 
 
@@ -1063,12 +1080,13 @@ class PmapExecutable:
                pci: ParallelCallableInfo,
                replicas: ReplicaInfo,
                parts: 'PartitionInfo',
-               shards: ShardInfo):
+               shards: ShardInfo,
+               tuple_args: bool):
     devices = pci.devices
     if devices is None:
       if shards.num_global_shards > xb.device_count(pci.backend):
         msg = ("compiling computation that requires {} logical devices, but only {} XLA "
-              "devices are available (num_replicas={}, num_partitions={})")
+               "devices are available (num_replicas={}, num_partitions={})")
         raise ValueError(msg.format(shards.num_global_shards,
                                     xb.device_count(pci.backend),
                                     replicas.num_global_replicas,
@@ -1124,7 +1142,7 @@ class PmapExecutable:
         device_assignment=device_assignment,
         use_spmd_partitioning=use_spmd_partitioning,
     )
-    compile_options.parameter_is_tupled_arguments = tuple_args(shards)
+    compile_options.parameter_is_tupled_arguments = tuple_args
 
     local_arg_parts_ = parts.local_arg_parts or [None] * len(pci.avals)
     input_sharding_specs = [
@@ -1168,8 +1186,10 @@ class PmapExecutable:
       # TODO(frostig): need `compile_replicated` to give us the XLA executable
       return PmapExecutable(None, execute_fun, None, pci.avals)
 
-    compiled = dispatch.compile_or_get_cached(
-        pci.backend, xla_computation, compile_options)
+    with dispatch.log_elapsed_time(
+        f"Finished XLA compilation of {pci.name} in {{elapsed_time}} sec"):
+      compiled = dispatch.compile_or_get_cached(
+          pci.backend, xla_computation, compile_options)
     handle_args = InputsHandler(
         compiled.local_devices(), input_sharding_specs, input_indices)
     execute_fun = partial(
@@ -1372,7 +1392,7 @@ def local_avals_to_results_handler(
 def global_avals_to_results_handler(global_out_avals: Sequence[ShapedArray],
                                     out_axes: Sequence[ArrayMapping],
                                     global_mesh):
-  if config.jax_gsda_out:
+  if config.jax_parallel_functions_output_gda:
     global_sharding_spec = mesh_sharding_specs(global_mesh.shape, global_mesh.axis_names)
     global_out_specs = [global_sharding_spec(aval, oa)
                         for aval, oa in safe_zip(global_out_avals, out_axes)]
@@ -1684,8 +1704,12 @@ def _mhlo_unshard(aval, axis_env, out_axis, xs, platform):
       transposed_dims = list(dims)
       transposed_dims.insert(out_axis, axis_env.sizes[-1])
       aval = aval.update(shape=transposed_dims)
-      out = mhlo.TransposeOp(mlir.aval_to_ir_type(aval), out,
-                             mlir.dense_int_elements(perm)).result
+      if _xla_extension_version < 49:
+        out = mhlo.TransposeOp(
+            mlir.aval_to_ir_type(aval), out,
+            mlir.dense_int_elements(perm)).result
+      else:
+        out = mhlo.TransposeOp(out, mlir.dense_int_elements(perm)).result
 
     # TODO(mattjj): remove this logic when AllReduce PRED supported on CPU / GPU
     if convert_bool:
@@ -1699,19 +1723,20 @@ def _mhlo_unshard(aval, axis_env, out_axis, xs, platform):
     raise TypeError(aval)
 
 
-def _pmap_lowering(ctx, avals_in, avals_out, *in_nodes, axis_name,
+def _pmap_lowering(ctx, *in_nodes, axis_name,
                    axis_size, global_axis_size, devices, name,
                    call_jaxpr, backend=None, in_axes, out_axes,
                    donated_invars, global_arg_shapes):
   del donated_invars  # Unused.
-  xla.check_backend_matches(backend, ctx.platform)
+  xla.check_backend_matches(backend, ctx.module_context.platform)
   # We in-line here rather than generating a Call HLO as in the xla_call
   # translation rule just because the extra tuple stuff is a pain.
-  if ctx.axis_env.names and devices is not None:
+  if ctx.module_context.axis_env.names and devices is not None:
     raise ValueError("Nested pmap with explicit devices argument.")
   if global_axis_size is None:
     global_axis_size = axis_size
-  new_env = xla.extend_axis_env(ctx.axis_env, axis_name, global_axis_size)
+  new_env = xla.extend_axis_env(ctx.module_context.axis_env, axis_name,
+                                global_axis_size)
   # Shard the in_nodes that are mapped
   in_avals = [v.aval for v in call_jaxpr.invars]
   in_nodes_sharded = (
@@ -1720,14 +1745,15 @@ def _pmap_lowering(ctx, avals_in, avals_out, *in_nodes, axis_name,
     for aval, in_node, in_axis in zip(in_avals, in_nodes, in_axes))
 
   with maybe_extend_axis_env(axis_name, global_axis_size, None):  # type: ignore
-    sub_ctx = ctx.replace(
+    sub_ctx = ctx.module_context.replace(
         axis_env=new_env,
-        name_stack=xla.extend_name_stack(ctx.name_stack,
+        name_stack=xla.extend_name_stack(ctx.module_context.name_stack,
                                          util.wrap_name(name, 'pmap')))
     sharded_outs = mlir.jaxpr_subcomp(sub_ctx, call_jaxpr, (),
                                       *in_nodes_sharded)
   out_avals = [v.aval for v in call_jaxpr.outvars]
-  outs = [_mhlo_unshard(aval, new_env, out_axis, shard, platform=ctx.platform)
+  outs = [_mhlo_unshard(aval, new_env, out_axis, shard,
+                        platform=ctx.module_context.platform)
           for aval, out_axis, shard in zip(out_avals, out_axes, sharded_outs)]
   return outs
 
@@ -1781,9 +1807,11 @@ class Mesh:
 
   @maybe_cached_property
   def local_mesh(self):
+    return self._local_mesh(xb.process_index())
+
+  def _local_mesh(self, process_index):
     if self.empty:
       return self
-    process_index = xb.process_index()
     is_local_device = np.vectorize(
         lambda d: d.process_index == process_index, otypes=[bool])(self.devices)
     subcube_indices = []
@@ -1899,6 +1927,7 @@ def lower_mesh_computation(
     in_is_gda: Sequence[bool]):
   assert not mesh.empty
   backend = xb.get_device_backend(mesh.devices.flat[0])
+  name_stack = extend_name_stack(wrap_name(transformed_name, 'xmap'))
 
   global_axis_sizes = mesh.shape
 
@@ -1923,7 +1952,9 @@ def lower_mesh_computation(
     assert tile_by_mesh_axes
     in_jaxpr_avals = in_tiled_avals
   with core.extend_axis_env_nd(mesh.shape.items()):
-    jaxpr, out_jaxpr_avals, consts = pe.trace_to_jaxpr_final(fun, in_jaxpr_avals)
+    with dispatch.log_elapsed_time(f"Finished tracing + transforming {name_stack} "
+                                   "in {elapsed_time} sec"):
+      jaxpr, out_jaxpr_avals, consts = pe.trace_to_jaxpr_final(fun, in_jaxpr_avals)
   if callable(out_axes):
     out_axes = out_axes()
   assert len(out_axes) == len(out_jaxpr_avals)
@@ -1962,43 +1993,57 @@ def lower_mesh_computation(
                            names=tuple(global_axis_sizes.keys()),
                            sizes=tuple(global_axis_sizes.values()))
   closed_jaxpr = core.ClosedJaxpr(jaxpr, consts)
-  name_stack = extend_name_stack(wrap_name(transformed_name, 'xmap'))
   module: Union[str, xc.XlaComputation]
+  module_name = f"xmap_{fun.__name__}"
   with core.extend_axis_env_nd(mesh.shape.items()):
     if config.jax_enable_mlir:
       module = mlir.lower_jaxpr_to_module(
-          closed_jaxpr, backend.platform, axis_env, name_stack, donated_invars,
-          replicated_args=replicated_args, arg_shardings=in_partitions,
-          result_shardings=out_partitions)
+          module_name, closed_jaxpr, backend.platform, axis_env, name_stack,
+          donated_invars, replicated_args=replicated_args,
+          arg_shardings=in_partitions, result_shardings=out_partitions)
     else:
       module = xla.lower_jaxpr_to_xla_module(
-          f"xmap_{fun.__name__}", closed_jaxpr, backend.platform, axis_env,
+          module_name, closed_jaxpr, backend.platform, axis_env,
           name_stack, tuple_args, donated_invars, replicated_args,
           in_partitions, out_partitions_t,
           partitions_are_protos=partitions_proto)
 
   return MeshComputation(
-      module, donated_invars, mesh, global_in_avals, global_out_avals,
-      in_axes, out_axes, spmd_lowering, tuple_args, in_is_gda)
+      name_stack, module, donated_invars, mesh=mesh, global_in_avals=global_in_avals,
+      global_out_avals=global_out_avals, in_axes=in_axes, out_axes=out_axes,
+      spmd_lowering=spmd_lowering, tuple_args=tuple_args, in_is_gda=in_is_gda)
 
 
 class MeshComputation:
-  def __init__(self, hlo, donated_invars, *compile_args):
-    self._executable = None
+  _executable: Optional['MeshExecutable']
+
+  def __init__(self, name: str, hlo: Union[str, xc.XlaComputation],
+               donated_invars: Sequence[bool], **compile_args):
+    self._name = name
     self._hlo = hlo
     self._donated_invars = donated_invars
     self.compile_args = compile_args
+    self._executable = None
 
   def hlo(self):
     # this is a method for api consistency with dispatch.XlaComputation
-    return self._hlo
+    if isinstance(self._hlo, xc.XlaComputation):
+      return self._hlo
+    return xe.mlir.mlir_module_to_xla_computation(
+        mlir.module_to_string(self._hlo),
+        use_tuple_args=self.compile_args["tuple_args"])
+
+  def mhlo(self) -> str:
+    if isinstance(self._hlo, xc.XlaComputation):
+      return xe.mlir.xla_computation_to_mlir_module(self._hlo)
+    return mlir.module_to_string(self._hlo)
 
   def compile(self,
               _allow_propagation_to_outputs : bool = False,
               _allow_compile_replicated : bool = True) -> 'MeshExecutable':
     if self._executable is None:
       self._executable = MeshExecutable.from_hlo(
-          self._hlo, *self.compile_args,
+          self._name, self._hlo, **self.compile_args,
           _allow_propagation_to_outputs=_allow_propagation_to_outputs,
           _allow_compile_replicated=_allow_compile_replicated)  # type: ignore
     return self._executable
@@ -2006,8 +2051,11 @@ class MeshComputation:
 
 def _get_input_metadata(global_in_avals, global_mesh, in_axes, in_is_gda):
   input_specs, input_indices, input_avals = [], [], []
+  num_local_devices = len(global_mesh.local_devices)
   for gaval, axis, is_gda in safe_zip(global_in_avals, in_axes, in_is_gda):
-    if is_gda:
+    # TODO(yashkatariya): Don't calculate input_indices and input_specs for GDA
+    # as GDA doesn't need it.
+    if is_gda or not axis:
       aval = gaval
       mesh = global_mesh
     else:
@@ -2016,7 +2064,12 @@ def _get_input_metadata(global_in_avals, global_mesh, in_axes, in_is_gda):
 
     spec = (mesh_sharding_specs(mesh.shape, mesh.axis_names)(aval, axis)
             if aval is not core.abstract_unit else None)
-    index = spec_to_indices(aval.shape, spec) if spec is not None else None
+    # We special case this logic to support fully replicated non-GDA values
+    # with non-contiguous submeshes
+    if not axis and not is_gda:
+      index = tuple((slice(None),) * aval.ndim for _ in range(num_local_devices))
+    else:
+      index = spec_to_indices(aval.shape, spec) if spec is not None else None
     input_specs.append(spec)
     input_indices.append(index)
     input_avals.append(aval)
@@ -2034,7 +2087,8 @@ class MeshExecutable:
     self._input_avals = input_avals
 
   @staticmethod
-  def from_hlo(computation: xc.XlaComputation,
+  def from_hlo(name: str,
+               computation: Union[str, xc.XlaComputation],
                mesh: Mesh,
                global_in_avals: Sequence[ShapedArray],
                global_out_avals: Sequence[ShapedArray],
@@ -2043,7 +2097,7 @@ class MeshExecutable:
                spmd_lowering: bool, tuple_args: bool,
                in_is_gda: Sequence[bool],
                _allow_propagation_to_outputs: bool,
-               _allow_compile_replicated: bool):
+               _allow_compile_replicated: bool) -> 'MeshExecutable':
     assert not mesh.empty
     backend = xb.get_device_backend(mesh.devices.flat[0])
 
@@ -2075,7 +2129,9 @@ class MeshExecutable:
           handle_outs)
       xla_executable = None
     else:
-      compiled = dispatch.compile_or_get_cached(backend, computation, compile_options)
+      with dispatch.log_elapsed_time(f"Finished XLA compilation of {name} "
+                                     "in {elapsed_time} sec"):
+        compiled = dispatch.compile_or_get_cached(backend, computation, compile_options)
       handle_args = InputsHandler(compiled.local_devices(), input_specs, input_indices)
       unsafe_call = partial(execute_replicated, compiled, backend, handle_args, handle_outs)
       xla_executable = compiled
