@@ -15,28 +15,28 @@
 # flake8: noqa: F401
 
 from jax._src.scipy.linalg import (
-  block_diag,
-  cholesky,
-  cho_factor,
-  cho_solve,
-  det,
-  eigh,
-  eigh_tridiagonal,
-  expm,
-  expm_frechet,
-  inv,
-  lu,
-  lu_factor,
-  lu_solve,
-  polar,
-  qr,
-  solve,
-  solve_triangular,
-  svd,
-  tril,
-  triu,
+  block_diag as block_diag,
+  cholesky as cholesky,
+  cho_factor as cho_factor,
+  cho_solve as cho_solve,
+  det as det,
+  eigh as eigh,
+  eigh_tridiagonal as eigh_tridiagonal,
+  expm as expm,
+  expm_frechet as expm_frechet,
+  inv as inv,
+  lu as lu,
+  lu_factor as lu_factor,
+  lu_solve as lu_solve,
+  polar as polar,
+  qr as qr,
+  solve as solve,
+  solve_triangular as solve_triangular,
+  svd as svd,
+  tril as tril,
+  triu as triu,
 )
 
 from jax._src.lax.polar import (
-  polar_unitary
+  polar_unitary as polar_unitary,
 )
