@@ -15,6 +15,6 @@
 # flake8: noqa: F401
 
 from jax._src.scipy.stats.multivariate_normal import (
-  logpdf,
-  pdf,
-)
\ No newline at end of file
+  logpdf as logpdf,
+  pdf as pdf,
+)
