@@ -14,18 +14,17 @@
 
 # flake8: noqa: F401
 from jax._src.util import (
-  HashableFunction,
-  as_hashable_function,
-  cache,
-  partial,
-  safe_map,
-  safe_zip,
-  split_dict,
-  split_list,
-  split_merge,
-  subvals,
-  toposort,
-  unzip2,
-  wrap_name,
-  wraps,
+  HashableFunction as HashableFunction,
+  as_hashable_function as as_hashable_function,
+  cache as cache,
+  safe_map as safe_map,
+  safe_zip as safe_zip,
+  split_dict as split_dict,
+  split_list as split_list,
+  split_merge as split_merge,
+  subvals as subvals,
+  toposort as toposort,
+  unzip2 as unzip2,
+  wrap_name as wrap_name,
+  wraps as wraps,
 )
